"""ThreatConnect Batch Import Module."""
# standard library
import gzip
import hashlib
import json
import math
import os
import re
import shelve
import sys
import threading
import time
import traceback
import uuid
from collections import deque
from typing import Any, Callable, Optional, Tuple, Union

from .group import (
    Adversary,
    Campaign,
    Document,
    Email,
    Event,
    Group,
    Incident,
    IntrusionSet,
    Report,
    Signature,
    Threat,
)
from .indicator import (
    ASN,
    CIDR,
    URL,
    Address,
    EmailAddress,
    File,
    Host,
    Indicator,
    Mutex,
    RegistryKey,
    UserAgent,
    custom_indicator_class_factory,
)

# import local modules for dynamic reference
module = __import__(__name__)


class Batch:
    """ThreatConnect Batch Import Module"""

    def __init__(
        self,
        tcex: object,
        owner: str,
        action: Optional[str] = 'Create',
        attribute_write_type: Optional[str] = 'Replace',
        halt_on_error: Optional[bool] = True,
        playbook_triggers_enabled: Optional[bool] = False,
    ):
        """Initialize Class properties.

        Args:
            tcex: An instance of TcEx object.
            owner: The ThreatConnect owner for Batch action.
            action: Action for the batch job ['Create', 'Delete'].
            attribute_write_type: Write type for Indicator attributes ['Append', 'Replace'].
            halt_on_error: If True any batch error will halt the batch job.
            playbook_triggers_enabled: **DEPRECATED**
        """
        self.tcex = tcex
        self._action = action
        self._attribute_write_type = attribute_write_type
        self._halt_on_error = halt_on_error
        self._owner = owner
        self._playbook_triggers_enabled = playbook_triggers_enabled

        # properties
        self._batch_max_chunk = 5000
        self._batch_max_size = 75_000_000  # max size in bytes
        self._file_merge_mode = None
        self._file_threads = []
        self._hash_collision_mode = None
        self._submit_thread = None

        # shelf settings
        self._group_shelf_fqfn = None
        self._indicator_shelf_fqfn = None

        # global overrides on batch/file errors
        self._halt_on_batch_error = None
        self._halt_on_file_error = None
        self._halt_on_poll_error = None

        # debug/saved flags
        self._saved_xids = None
        self._saved_groups = None  # indicates groups shelf file was provided
        self._saved_indicators = None  # indicates indicators shelf file was provided
        self.enable_saved_file = False

        # default properties
        self._batch_data_count = None
        self._poll_interval = None
        self._poll_interval_times = []
        self._poll_timeout = 3600

        # containers
        self._groups = None
        self._groups_shelf = None
        self._indicators = None
        self._indicators_shelf = None

        # build custom indicator classes
        self._gen_indicator_class()

        # batch debug/replay variables
        self._debug = None
        self.debug_path = os.path.join(self.tcex.args.tc_temp_path, 'DEBUG')
        self.debug_path_batch = os.path.join(self.debug_path, 'batch_data')
        self.debug_path_group_shelf = os.path.join(self.debug_path, 'groups-saved')
        self.debug_path_indicator_shelf = os.path.join(self.debug_path, 'indicators-saved')
        self.debug_path_files = os.path.join(self.debug_path, 'batch_files')
        self.debug_path_xids = os.path.join(self.debug_path, 'xids-saved')

    @property
    def _critical_failures(self):  # pragma: no cover
        """Return Batch critical failure messages."""
        return [
            'Encountered an unexpected Exception while processing batch job',
            'would exceed the number of allowed indicators',
        ]

    def _gen_indicator_class(self):  # pragma: no cover
        """Generate Custom Indicator Classes."""
        for entry in self.tcex.indicator_types_data.values():
            name = entry.get('name')
            class_name = name.replace(' ', '')
            # temp fix for API issue where boolean are returned as strings
            entry['custom'] = self.tcex.utils.to_bool(entry.get('custom'))

            if class_name in globals():
                # skip Indicator Type if a class already exists
                continue

            # Custom Indicator can have 3 values. Only add the value if it is set.
            value_fields = []
            if entry.get('value1Label'):
                value_fields.append(entry['value1Label'])
            if entry.get('value2Label'):
                value_fields.append(entry['value2Label'])
            if entry.get('value3Label'):
                value_fields.append(entry['value3Label'])
            value_count = len(value_fields)

            class_data = {}
            # Add Class for each Custom Indicator type to this module
            custom_class = custom_indicator_class_factory(name, Indicator, class_data, value_fields)
            setattr(module, class_name, custom_class)

            # Add Custom Indicator Method
            self._gen_indicator_method(name, custom_class, value_count)

    def _gen_indicator_method(
        self, name: str, custom_class: object, value_count: int
    ) -> None:  # pragma: no cover
        """Dynamically generate custom Indicator methods.

        Args:
            name (str): The name of the method.
            custom_class (object): The class to add.
            value_count (int): The number of value parameters to support.
        """
        method_name = name.replace(' ', '_').lower()

        # Add Method for each Custom Indicator class
        def method_1(value1: str, xid, **kwargs):  # pylint: disable=possibly-unused-variable
            """Add Custom Indicator data to Batch object"""
            indicator_obj = custom_class(value1, xid, **kwargs)
            return self._indicator(indicator_obj, kwargs.get('store', True))

        def method_2(
            value1: str, value2: str, xid, **kwargs
        ):  # pylint: disable=possibly-unused-variable
            """Add Custom Indicator data to Batch object"""
            indicator_obj = custom_class(value1, value2, xid, **kwargs)
            return self._indicator(indicator_obj, kwargs.get('store', True))

        def method_3(
            value1: str, value2: str, value3: str, xid, **kwargs
        ):  # pylint: disable=possibly-unused-variable
            """Add Custom Indicator data to Batch object"""
            indicator_obj = custom_class(value1, value2, value3, xid, **kwargs)
            return self._indicator(indicator_obj, kwargs.get('store', True))

        method = locals()[f'method_{value_count}']
        setattr(self, method_name, method)

    def _group(
        self, group_data: Union[dict, object], store: Optional[bool] = True
    ) -> Union[dict, object]:
        """Return previously stored group or new group.

        Args:
            group_data: An Group dict or instance of Group object.
            store: If True the group data will be stored in instance list.

        Returns:
            Union[dict, object]: The new Group dict/object or the previously stored dict/object.
        """
        if store is False:
            return group_data

        if isinstance(group_data, dict):
            # get xid from dict
            xid = group_data.get('xid')
        else:
            # get xid from object
            xid = group_data.xid

        if self.groups.get(xid) is not None:
            # return existing group from memory
            group_data = self.groups.get(xid)
        elif self.groups_shelf.get(xid) is not None:
            # return existing group from shelf
            group_data = self.groups_shelf.get(xid)
        else:
            # store new group
            self.groups[xid] = group_data
        return group_data

    def _indicator(
        self, indicator_data: Union[dict, object], store: Optional[bool] = True
    ) -> Union[dict, object]:
        """Return previously stored indicator or new indicator.

        Args:
            indicator_data: An Indicator dict or instance of Indicator object.
            store: If True the indicator data will be stored in instance list.

        Returns:
            Union[dict, object]: The new Indicator dict/object or the previously stored dict/object.
        """
        if store is False:
            return indicator_data

        if isinstance(indicator_data, dict):
            # get xid from dict
            xid = indicator_data.get('xid')
        else:
            # get xid from object
            xid = indicator_data.xid

        if self.indicators.get(xid) is not None:
            # return existing indicator from memory
            indicator_data = self.indicators.get(xid)
        elif self.indicators_shelf.get(xid) is not None:
            # return existing indicator from shelf
            indicator_data = self.indicators_shelf.get(xid)
        else:
            # store new indicators
            self.indicators[xid] = indicator_data
        return indicator_data

    @staticmethod
    def _indicator_values(indicator: str) -> list:
        """Process indicators expanding file hashes/custom indicators into multiple entries.

        Args:
            indicator: Indicator value represented as " : " delimited string.

        Returns:
            list: The list of indicators split on " : ".
        """
        indicator_list = [indicator]
        if indicator.count(' : ') > 0:
            # handle all multi-valued indicators types (file hashes and custom indicators)
            indicator_list = []

            # group 1 - lazy capture everything to first <space>:<space> or end of line
            iregx_pattern = r'^(.*?(?=\s\:\s|$))?'
            iregx_pattern += r'(?:\s\:\s)?'  # remove <space>:<space>
            # group 2 - look behind for <space>:<space>, lazy capture everything
            #           to look ahead (optional <space>):<space> or end of line
            iregx_pattern += r'((?<=\s\:\s).*?(?=(?:\s)?\:\s|$))?'
            iregx_pattern += r'(?:(?:\s)?\:\s)?'  # remove (optional <space>):<space>
            # group 3 - look behind for <space>:<space>, lazy capture everything
            #           to look ahead end of line
            iregx_pattern += r'((?<=\s\:\s).*?(?=$))?$'
            iregx = re.compile(iregx_pattern)

            indicators = iregx.search(indicator)
            if indicators is not None:
                indicator_list = list(indicators.groups())

        return indicator_list

    @property
    def action(self):
        """Return batch action."""
        return self._action

    @action.setter
    def action(self, action):
        """Set batch action."""
        self._action = action

    def add_group(self, group_data: dict, **kwargs) -> Union[dict, object]:
        """Add a group to Batch Job.

        .. code-block:: javascript

            {
                "name": "Example Incident",
                "type": "Incident",
                "attribute": [{
                    "type": "Description",
                    "displayed": false,
                    "value": "Example Description"
                }],
                "xid": "e336e2dd-5dfb-48cd-a33a-f8809e83e904",
                "associatedGroupXid": [
                    "e336e2dd-5dfb-48cd-a33a-f8809e83e904:58",
                ],
                "tag": [{
                    "name": "China"
                }]
            }

        Args:
            group_data: The full Group data including attributes, labels, tags, and
                associations.
            store: (bool, kwargs): Advanced - Defaults to True. If True
                the indicator data will be stored in instance list.

        Returns:
            Union[dict, object]: The new group dict/object or the previously stored dict/object.
        """
        return self._group(group_data, kwargs.get('store', True))

    def add_indicator(self, indicator_data: dict, **kwargs) -> Union[dict, object]:
        """Add an indicator to Batch Job.

        .. code-block:: javascript

            {
                "type": "File",
                "rating": 5.00,
                "confidence": 50,
                "summary": "53c3609411c83f363e051d455ade78a7
                            : 57a49b478310e4313c54c0fee46e4d70a73dd580
                            : db31cb2a748b7e0046d8c97a32a7eb4efde32a0593e5dbd58e07a3b4ae6bf3d7",
                "associatedGroups": [
                    {
                        "groupXid": "e336e2dd-5dfb-48cd-a33a-f8809e83e904"
                    }
                ],
                "attribute": [{
                    "type": "Source",
                    "displayed": true,
                    "value": "Malware Analysis provided by external AMA."
                }],
                "fileOccurrence": [{
                    "fileName": "drop1.exe",
                    "date": "2017-03-03T18:00:00-06:00"
                }],
                "tag": [{
                    "name": "China"
                }],
                "xid": "e336e2dd-5dfb-48cd-a33a-f8809e83e904:170139"
            }

        Args:
            indicator_data: The Full Indicator data including attributes, labels, tags,
                and associations.
            store: (bool, kwargs): Advanced - Defaults to True. If True
                the indicator data will be stored in instance list.

        Returns:
            Union[dict, object]: The new group dict/object or the previously stored dict/object.
        """
        if indicator_data.get('type') not in ['Address', 'EmailAddress', 'File', 'Host', 'URL']:
            # for custom indicator types the valueX fields are required.
            # using the summary we can build the values
            index = 1
            for value in self._indicator_values(indicator_data.get('summary')):
                indicator_data[f'value{index}'] = value
                index += 1
        if indicator_data.get('type') == 'File':
            # convert custom field name to the appropriate value for batch v2
            size = indicator_data.pop('size', None)
            if size is not None:
                indicator_data['intValue1'] = size
        if indicator_data.get('type') == 'Host':
            # convert custom field name to the appropriate value for batch v2
            dns_active = indicator_data.pop('dnsActive', None)
            if dns_active is not None:
                indicator_data['flag1'] = dns_active
            whois_active = indicator_data.pop('whoisActive', None)
            if whois_active is not None:
                indicator_data['flag2'] = whois_active
        return self._indicator(indicator_data, kwargs.get('store', True))

    def address(self, ip: str, **kwargs) -> Address:
        """Add Address data to Batch object.

        Args:
            ip: The value for this Indicator.
            confidence (str, kwargs): The threat confidence for this Indicator.
            date_added (str, kwargs): The date timestamp the Indicator was created.
            last_modified (str, kwargs): The date timestamp the Indicator was last modified.
            rating (str, kwargs): The threat rating for this Indicator.
            xid (str, kwargs): The external id for this Indicator.
            store: (bool, kwargs): Advanced - Defaults to True. If True
                the indicator data will be stored in instance list.

        Returns:
            Address: An instance of the Address class.
        """
        indicator_obj = Address(ip, **kwargs)
        return self._indicator(indicator_obj, kwargs.get('store', True))

    def adversary(self, name: str, **kwargs) -> Adversary:
        """Add Adversary data to Batch object.

        Args:
            name: The name for this Group.
            date_added (str, kwargs): The date timestamp the Indicator was created.
            xid (str, kwargs): The external id for this Group.
            store: (bool, kwargs): Advanced - Defaults to True. If True
                the indicator data will be stored in instance list.

        Returns:
            Adversary: An instance of the Adversary class.
        """
        group_obj = Adversary(name, **kwargs)
        return self._group(group_obj, kwargs.get('store', True))

    def asn(self, as_number: str, **kwargs) -> ASN:
        """Add ASN data to Batch object.

        Args:
            as_number: The value for this Indicator.
            confidence (str, kwargs): The threat confidence for this Indicator.
            date_added (str, kwargs): The date timestamp the Indicator was created.
            last_modified (str, kwargs): The date timestamp the Indicator was last modified.
            rating (str, kwargs): The threat rating for this Indicator.
            xid (str, kwargs): The external id for this Indicator.
            store: (bool, kwargs): Advanced - Defaults to True. If True
                the indicator data will be stored in instance list.

        Returns:
            ASN: An instance of the ASN class.
        """
        indicator_obj = ASN(as_number, **kwargs)
        return self._indicator(indicator_obj, kwargs.get('store', True))

    @property
    def attribute_write_type(self):
        """Return batch attribute write type."""
        return self._attribute_write_type

    @attribute_write_type.setter
    def attribute_write_type(self, attribute_write_type: str):
        """Set batch attribute write type."""
        self._attribute_write_type = attribute_write_type

    def campaign(self, name: str, **kwargs) -> Campaign:
        """Add Campaign data to Batch object.

        Args:
            name: The name for this Group.
            date_added (str, kwargs): The date timestamp the Indicator was created.
            first_seen (str, kwargs): The first seen datetime expression for this Group.
            xid (str, kwargs): The external id for this Group.
            store: (bool, kwargs): Advanced - Defaults to True. If True
                the indicator data will be stored in instance list.

        Returns:
            Campaign: An instance of the Campaign class.
        """
        group_obj = Campaign(name, **kwargs)
        return self._group(group_obj, kwargs.get('store', True))

    def cidr(self, block: str, **kwargs) -> CIDR:
        """Add CIDR data to Batch object.

        Args:
            block: The value for this Indicator.
            confidence (str, kwargs): The threat confidence for this Indicator.
            date_added (str, kwargs): The date timestamp the Indicator was created.
            last_modified (str, kwargs): The date timestamp the Indicator was last modified.
            rating (str, kwargs): The threat rating for this Indicator.
            xid (str, kwargs): The external id for this Indicator.
            store: (bool, kwargs): Advanced - Defaults to True. If True
                the indicator data will be stored in instance list.

        Returns:
            CIDR: An instance of the CIDR class.
        """
        indicator_obj = CIDR(block, **kwargs)
        return self._indicator(indicator_obj, kwargs.get('store', True))

    def close(self) -> None:
        """Cleanup batch job."""
        # allow pol thread to complete before wrapping up
        if hasattr(self._submit_thread, 'is_alive'):
            self._submit_thread.join()

        # allow file threads to complete before wrapping up job
        for t in self._file_threads:
            t.join()

        self.groups_shelf.close()
        self.indicators_shelf.close()
        if not self.debug and not self.enable_saved_file:
            # delete saved files
            if os.path.isfile(self.group_shelf_fqfn):
                os.remove(self.group_shelf_fqfn)
            if os.path.isfile(self.group_shelf_fqfn):
                os.remove(self.indicator_shelf_fqfn)

    @property
    def data(self):
        """Return the batch indicator/group and file data to be sent to the ThreatConnect API.

        **Processing Order:**
        * Process groups in memory up to max batch size.
        * Process groups in shelf to max batch size.
        * Process indicators in memory up to max batch size.
        * Process indicators in shelf up to max batch size.

        This method will remove the group/indicator from memory and/or shelf.

        Returns:
            dict: A dictionary of group, indicators, and/or file data.
        """
        entity_count = 0
        entity_size = 0
        data = {'file': {}, 'group': [], 'indicator': []}

        # process group from memory, returning if max values have been reached
        if self.data_groups(data, self.groups, entity_count, entity_size) is True:
            return data

        # process group from shelf file, returning if max values have been reached
        if self.data_groups(data, self.groups_shelf, entity_count, entity_size) is True:
            return data

        # process indicator from memory, returning if max values have been reached
        if self.data_indicators(data, self.indicators, entity_count, entity_size) is True:
            return data

        # process indicator from shelf file, returning if max values have been reached
        if self.data_indicators(data, self.indicators_shelf, entity_count, entity_size) is True:
            return data

        return data

    def data_group_association(
        self, data: dict, entity_count: int, entity_size: int, xid: str
    ) -> Tuple[int, int]:
        """Return group dict array following all associations.

        The *data* dict is passed by reference to make it easier to update both the group data
        and file data inline versus passing the data all the way back up to the calling methods.
        The entity_count and entity size values are immutable and must be returned.

        Args:
            data: The data dict to update with group and file data.
            entity_count: The total count of all entities collected.
            entity_size: The total size in bytes of all entities collected.
            xid: The xid of the group to retrieve associations.

        Returns:
            Tuple[int, int]: Return tuple containting updated entity_count and entity_size.
        """
        xids = deque()
        xids.append(xid)

        while xids:
            xid = xids.popleft()  # remove current xid
            group_data = None

            if xid in self.groups:
                group_data = self.groups.get(xid)
                del self.groups[xid]
            elif xid in self.groups_shelf:
                group_data = self.groups_shelf.get(xid)
                del self.groups_shelf[xid]

            if group_data:
                file_data, group_data = self.data_group_type(group_data)
                data['group'].append(group_data)
                if file_data:
                    data['file'][xid] = file_data

                # update entity trackers
                entity_count += 1
                entity_size += sys.getsizeof(json.dumps(group_data))

                # groups.append(group_data)
                xids.extend(group_data.get('associatedGroupXid', []))

        return entity_count, entity_size

    @staticmethod
    def data_group_type(group_data: Union[dict, object]) -> Tuple[dict, dict]:
        """Return dict representation of group data and file data.

        Args:
            group_data: The group data dict or object.

        Returns:
            Tuple[dict, dict]: A tuple containing file_data and group_data.
        """
        file_data = {}
        if isinstance(group_data, dict):
            # process file content
            file_content = group_data.pop('fileContent', None)
            if file_content is not None:
                file_data = {
                    'fileContent': file_content,
                    'fileName': group_data.get('fileName'),
                    'type': group_data.get('type'),
                }
        else:
            # get the file data from the object and return dict format of object
            if group_data.data.get('type') in ['Document', 'Report']:
                file_data = group_data.file_data
            group_data = group_data.data

        return file_data, group_data

    def data_groups(self, data: dict, groups: list, entity_count: int, entity_size: int) -> bool:
        """Process Group data.

        Args:
            data: The data dict to update with group and file data.
            groups: The list of groups to process.
            entity_count: The total count of all entities collected.
            entity_size: The total size in bytes of all entities collected.

        Returns:
            bool: True if max values have been hit, else False.
        """
        # convert groups.keys() to a list to prevent dictionary change error caused by
        # the data_group_association function deleting items from the object.

        # process group objects
        for xid in list(groups.keys()):
            # get association from group data
            entity_count, entity_size = self.data_group_association(
                data, entity_count, entity_size, xid
            )

            if entity_count % 2_500 == 0:
                # log count/size at a sane level
                self.tcex.log.info(
                    'feature=batch, action=data-groups, '
                    f'count={entity_count:,}, size={entity_size:,}'
                )

            if entity_count >= self._batch_max_chunk or entity_size >= self._batch_max_size:
                # stop processing xid once max limit are reached
                self.tcex.log.info(
                    f'feature=batch, event=max-hit, count={entity_count:,}, size={entity_size:,}'
                )
                return True
        return False

    def data_indicators(
        self, data: dict, indicators: list, entity_count: int, entity_size: int
    ) -> bool:
        """Process Indicator data.

        Args:
            data: The data dict to update with group and file data.
            indicators: The list of indicators to process.
            entity_count: The total count of all entities collected.
            entity_size: The total size in bytes of all entities collected.

        Returns:
            bool: True if max values have been hit, else False.
        """
        # process indicator objects
        for xid, indicator_data in list(indicators.items()):
            if not isinstance(indicator_data, dict):
                indicator_data = indicator_data.data
            data['indicator'].append(indicator_data)
            del indicators[xid]

            # update entity trackers
            entity_count += 1
            entity_size += sys.getsizeof(json.dumps(indicator_data))

            if entity_count % 2_500 == 0:
                # log count/size at a sane level
                self.tcex.log.info(
                    'feature=batch, action=data-indicator, '
                    f'count={entity_count:,}, size={entity_size:,}'
                )

            if entity_count >= self._batch_max_chunk or entity_size >= self._batch_max_size:
                # stop processing xid once max limit are reached
                self.tcex.log.info(
                    f'feature=batch, event=max-hit, count={entity_count:,}, size={entity_size:,}'
                )
                return True
        return False

    @property
    def debug(self):
        """Return debug setting"""
        if self._debug is None:
            self._debug = False

            # switching DEBUG file to a directory
            if os.path.isfile(self.debug_path):
                os.remove(self.debug_path)
                os.makedirs(self.debug_path, exist_ok=True)

            if os.path.isdir(self.debug_path) and os.access(self.debug_path, os.R_OK):
                # create directories only required when debug is enabled
                # batch_json - store the batch*.json files
                # documents - store the file downloads (e.g., *.pdf)
                # reports - store the file downloads (e.g., *.pdf)
                os.makedirs(self.debug_path, exist_ok=True)
                os.makedirs(self.debug_path_batch, exist_ok=True)
                os.makedirs(self.debug_path_files, exist_ok=True)
                self._debug = True
        return self._debug

    def document(self, name: str, file_name: str, **kwargs) -> Document:
        """Add Document data to Batch object.

        Args:
            name: The name for this Group.
            file_name: The name for the attached file for this Group.
            date_added (str, kwargs): The date timestamp the Indicator was created.
            file_content (str;method, kwargs): The file contents or
                callback method to retrieve file content.
            malware (bool, kwargs): If true the file is considered malware.
            password (bool, kwargs): If malware is true a password for the zip archive is
            xid (str, kwargs): The external id for this Group.
            store: (bool, kwargs): Advanced - Defaults to True. If True
                the indicator data will be stored in instance list.

        Returns:
            Document: An instance of the Document class.
        """
        group_obj = Document(name, file_name, **kwargs)
        return self._group(group_obj, kwargs.get('store', True))

    def email(self, name: str, subject: str, header: str, body: str, **kwargs) -> Email:
        """Add Email data to Batch object.

        Args:
            name: The name for this Group.
            subject: The subject for this Email.
            header: The header for this Email.
            body: The body for this Email.
            date_added (str, kwargs): The date timestamp the Indicator was created.
            from_addr (str, kwargs): The **from** address for this Email.
            to_addr (str, kwargs): The **to** address for this Email.
            xid (str, kwargs): The external id for this Group.
            store: (bool, kwargs): Advanced - Defaults to True. If True
                the indicator data will be stored in instance list.

        Returns:
            Email: An instance of the Email class.
        """
        group_obj = Email(name, subject, header, body, **kwargs)
        return self._group(group_obj, kwargs.get('store', True))

    def email_address(self, address: str, **kwargs) -> EmailAddress:
        """Add Email Address data to Batch object.

        Args:
            address: The value for this Indicator.
            confidence (str, kwargs): The threat confidence for this Indicator.
            date_added (str, kwargs): The date timestamp the Indicator was created.
            last_modified (str, kwargs): The date timestamp the Indicator was last modified.
            rating (str, kwargs): The threat rating for this Indicator.
            xid (str, kwargs): The external id for this Indicator.
            store: (bool, kwargs): Advanced - Defaults to True. If True
                the indicator data will be stored in instance list.

        Returns:
            EmailAddress: An instance of the EmailAddress class.
        """
        indicator_obj = EmailAddress(address, **kwargs)
        return self._indicator(indicator_obj, kwargs.get('store', True))

    @property
    def error_codes(self):
        """Return static list of Batch error codes and short description"""
        return {
            '0x1001': 'General Error',
            '0x1002': 'Permission Error',
            '0x1003': 'JsonSyntax Error',
            '0x1004': 'Internal Error',
            '0x1005': 'Invalid Indicator Error',
            '0x1006': 'Invalid Group Error',
            '0x1007': 'Item Not Found Error',
            '0x1008': 'Indicator Limit Error',
            '0x1009': 'Association Error',
            '0x100A': 'Duplicate Item Error',
            '0x100B': 'File IO Error',
            '0x2001': 'Indicator Partial Loss Error',
            '0x2002': 'Group Partial Loss Error',
            '0x2003': 'File Hash Merge Error',
        }

    def errors(self, batch_id: int, halt_on_error: Optional[bool] = True) -> list:
        """Retrieve Batch errors to ThreatConnect API.

        .. code-block:: javascript

            [{
                "errorReason": "Incident incident-001 has an invalid status.",
                "errorSource": "incident-001 is not valid."
            }, {
                "errorReason": "Incident incident-002 has an invalid status.",
                "errorSource":"incident-002 is not valid."
            }]

        Args:
            batch_id: The ID returned from the ThreatConnect API for the current batch job.
            halt_on_error: If True any exception will raise an error.

        Returns:
            list: A list of batch errors.
        """
        errors = []
        try:
            self.tcex.log.debug(f'feature=batch, event=retrieve-errors, batch-id={batch_id}')
            r = self.tcex.session.get(f'/v2/batch/{batch_id}/errors')
            # API does not return correct content type
            if r.ok:
                errors = json.loads(r.text)
            # temporarily process errors to find "critical" errors.
            # FR in core to return error codes.
            for error in errors:
                error_reason = error.get('errorReason')
                for error_msg in self._critical_failures:
                    if re.findall(error_msg, error_reason):
                        self.tcex.handle_error(10500, [error_reason], halt_on_error)
            return errors
        except Exception as e:
            self.tcex.handle_error(560, [e], halt_on_error)

    def event(self, name: str, **kwargs) -> Event:
        """Add Event data to Batch object.

        Args:
            name: The name for this Group.
            date_added (str, kwargs): The date timestamp the Indicator was created.
            event_date (str, kwargs): The event datetime expression for this Group.
            status (str, kwargs): The status for this Group.
            xid (str, kwargs): The external id for this Group.
            store: (bool, kwargs): Advanced - Defaults to True. If True
                the indicator data will be stored in instance list.

        Returns:
            Event: An instance of the Event class.
        """
        group_obj = Event(name, **kwargs)
        return self._group(group_obj, kwargs.get('store', True))

    def file(
        self,
        md5: Optional[str] = None,
        sha1: Optional[str] = None,
        sha256: Optional[str] = None,
        **kwargs,
    ) -> File:
        """Add File data to Batch object.

        .. note:: A least one file hash value must be specified.

        Args:
            md5: The md5 value for this Indicator.
            sha1: The sha1 value for this Indicator.
            sha256: The sha256 value for this Indicator.
            confidence (str, kwargs): The threat confidence for this Indicator.
            date_added (str, kwargs): The date timestamp the Indicator was created.
            last_modified (str, kwargs): The date timestamp the Indicator was last modified.
            rating (str, kwargs): The threat rating for this Indicator.
            size (str, kwargs): The file size for this Indicator.
            xid (str, kwargs): The external id for this Indicator.
            store: (bool, kwargs): Advanced - Defaults to True. If True
                the indicator data will be stored in instance list.

        Returns:
            File: An instance of the File class.

        """
        indicator_obj = File(md5, sha1, sha256, **kwargs)
        return self._indicator(indicator_obj, kwargs.get('store', True))

    def file_merge_mode(self, value: str) -> None:
        """Set the file merge mode for the entire batch job.

        Args:
            value: A value of Distribute or Merge.
        """
        self._file_merge_mode = value

    @staticmethod
    def generate_xid(identifier: Optional[Union[list, str]] = None):
        """Generate xid from provided identifiers.

        .. Important::  If no identifier is provided a unique xid will be returned, but it will
                        not be reproducible. If a list of identifiers are provided they must be
                        in the same order to generate a reproducible xid.

        Args:
            identifier:  Optional *string* value(s) to be
               used to make a unique and reproducible xid.

        """
        if identifier is None:
            identifier = str(uuid.uuid4())
        elif isinstance(identifier, list):
            identifier = '-'.join([str(i) for i in identifier])
            identifier = hashlib.sha256(identifier.encode('utf-8')).hexdigest()
        return hashlib.sha256(identifier.encode('utf-8')).hexdigest()

    def group(self, group_type: str, name: str, **kwargs) -> object:
        """Add Group data to Batch object.

        Args:
            group_type: The ThreatConnect define Group type.
            name: The name for this Group.
            date_added (str, kwargs): The date timestamp the Indicator was created.
            xid (str, kwargs): The external id for this Group.
            store: (bool, kwargs): Advanced - Defaults to True. If True
                the indicator data will be stored in instance list.

        Returns:
            object: An instance of one of the Group classes.
        """
        group_obj = Group(group_type, name, **kwargs)
        return self._group(group_obj, kwargs.get('store', True))

    @property
    def group_shelf_fqfn(self):
        """Return groups shelf fully qualified filename.

        For testing/debugging a previous shelf file can be copied into the tc_temp_path directory
        instead of creating a new shelf file.
        """
        if self._group_shelf_fqfn is None:
            # new shelf file
            self._group_shelf_fqfn = os.path.join(
                self.tcex.args.tc_temp_path, f'groups-{str(uuid.uuid4())}'
            )

            # saved shelf file
            if self.saved_groups:
                self._group_shelf_fqfn = self.debug_path_group_shelf
        return self._group_shelf_fqfn

    @property
    def groups(self) -> dict:
        """Return dictionary of all Groups data."""
        if self._groups is None:
            # plain dict, but could be something else in future
            self._groups = {}
        return self._groups

    @property
    def groups_shelf(self) -> object:
        """Return dictionary of all Groups data."""
        if self._groups_shelf is None:
            self._groups_shelf = shelve.open(self.group_shelf_fqfn, writeback=False)
        return self._groups_shelf

    @property
    def halt_on_error(self) -> bool:
        """Return batch halt on error setting."""
        return self._halt_on_error

    @halt_on_error.setter
    def halt_on_error(self, halt_on_error: bool):
        """Set batch halt on error setting."""
        self._halt_on_error = halt_on_error

    @property
    def halt_on_batch_error(self) -> bool:
        """Return halt on batch error value."""
        return self._halt_on_batch_error

    @halt_on_batch_error.setter
    def halt_on_batch_error(self, value: bool):
        """Set batch halt on batch error value."""
        if isinstance(value, bool):
            self._halt_on_batch_error = value

    @property
    def halt_on_file_error(self) -> bool:
        """Return halt on file post error value."""
        return self._halt_on_file_error

    @halt_on_file_error.setter
    def halt_on_file_error(self, value: bool):
        """Set halt on file post error value."""
        if isinstance(value, bool):
            self._halt_on_file_error = value

    @property
    def halt_on_poll_error(self) -> bool:
        """Return halt on poll error value."""
        return self._halt_on_poll_error

    @halt_on_poll_error.setter
    def halt_on_poll_error(self, value: bool):
        """Set batch halt on poll error value."""
        if isinstance(value, bool):
            self._halt_on_poll_error = value

    def hash_collision_mode(self, value: str):
        """Set the file hash collision mode for the entire batch job.

        Args:
            value: A value of Split, IgnoreIncoming, IgnoreExisting, FavorIncoming,
                and FavorExisting.
        """
        self._hash_collision_mode = value

    def host(self, hostname: str, **kwargs) -> Host:
        """Add Host data to Batch object.

        Args:
            hostname: The value for this Indicator.
            confidence (str, kwargs): The threat confidence for this Indicator.
            date_added (str, kwargs): The date timestamp the Indicator was created.
            dns_active (bool, kwargs): If True DNS active is enabled for this indicator.
            last_modified (str, kwargs): The date timestamp the Indicator was last modified.
            rating (str, kwargs): The threat rating for this Indicator.
            whois_active (bool, kwargs): If True WhoIs active is enabled for this indicator.
            xid (str, kwargs): The external id for this Indicator.
            store: (bool, kwargs): Advanced - Defaults to True. If True
                the indicator data will be stored in instance list.

        Returns:
            Host: An instance of the Host class.
        """
        indicator_obj = Host(hostname, **kwargs)
        return self._indicator(indicator_obj, kwargs.get('store', True))

    def incident(self, name: str, **kwargs) -> Incident:
        """Add Incident data to Batch object.

        Args:
            name: The name for this Group.
            date_added (str, kwargs): The date timestamp the Indicator was created.
            event_date (str, kwargs): The event datetime expression for this Group.
            status (str, kwargs): The status for this Group.
            xid (str, kwargs): The external id for this Group.
            store: (bool, kwargs): Advanced - Defaults to True. If True
                the indicator data will be stored in instance list.

        Returns:
            Incident: An instance of the Incident class.
        """
        group_obj = Incident(name, **kwargs)
        return self._group(group_obj, kwargs.get('store', True))

    def indicator(self, indicator_type: str, summary: str, **kwargs) -> object:
        """Add Indicator data to Batch object.

        Args:
            indicator_type: The ThreatConnect define Indicator type.
            summary: The value for this Indicator.
            confidence (str, kwargs): The threat confidence for this Indicator.
            date_added (str, kwargs): The date timestamp the Indicator was created.
            last_modified (str, kwargs): The date timestamp the Indicator was last modified.
            rating (str, kwargs): The threat rating for this Indicator.
            xid (str, kwargs): The external id for this Indicator.
            store: (bool, kwargs): Advanced - Defaults to True. If True
                the indicator data will be stored in instance list.

        Returns:
            object: An instance of one of the Indicator classes.
        """
        indicator_obj = Indicator(indicator_type, summary, **kwargs)
        return self._indicator(indicator_obj, kwargs.get('store', True))

    @property
    def indicator_shelf_fqfn(self) -> str:
        """Return indicator shelf fully qualified filename.

        For testing/debugging a previous shelf file can be copied into the tc_temp_path directory
        instead of creating a new shelf file.
        """
        if self._indicator_shelf_fqfn is None:
            # new shelf file
            self._indicator_shelf_fqfn = os.path.join(
                self.tcex.args.tc_temp_path, f'indicators-{str(uuid.uuid4())}'
            )

            # saved shelf file
            if self.saved_indicators:
                self._indicator_shelf_fqfn = self.debug_path_indicator_shelf
        return self._indicator_shelf_fqfn

    @property
    def indicators(self) -> dict:
        """Return dictionary of all Indicator data."""
        if self._indicators is None:
            # plain dict, but could be something else in future
            self._indicators = {}
        return self._indicators

    @property
    def indicators_shelf(self) -> object:
        """Return dictionary of all Indicator data."""
        if self._indicators_shelf is None:
            self._indicators_shelf = shelve.open(self.indicator_shelf_fqfn, writeback=False)
        return self._indicators_shelf

    def intrusion_set(self, name: str, **kwargs) -> IntrusionSet:
        """Add Intrusion Set data to Batch object.

        Args:
            name: The name for this Group.
            date_added (str, kwargs): The date timestamp the Indicator was created.
            xid (str, kwargs): The external id for this Group.
            store: (bool, kwargs): Advanced - Defaults to True. If True
                the indicator data will be stored in instance list.

        Returns:
            IntrusionSet: An instance of the IntrusionSet class.
        """
        group_obj = IntrusionSet(name, **kwargs)
        return self._group(group_obj, kwargs.get('store', True))

    def mutex(self, mutex: str, **kwargs) -> Mutex:
        """Add Mutex data to Batch object.

        Args:
            mutex: The value for this Indicator.
            confidence (str, kwargs): The threat confidence for this Indicator.
            date_added (str, kwargs): The date timestamp the Indicator was created.
            last_modified (str, kwargs): The date timestamp the Indicator was last modified.
            rating (str, kwargs): The threat rating for this Indicator.
            xid (str, kwargs): The external id for this Indicator.
            store: (bool, kwargs): Advanced - Defaults to True. If True
                the indicator data will be stored in instance list.

        Returns:
            Mutex: An instance of the Mutex class.
        """
        indicator_obj = Mutex(mutex, **kwargs)
        return self._indicator(indicator_obj, kwargs.get('store', True))

    def poll(
        self,
        batch_id: int,
        retry_seconds: Optional[int] = None,
        back_off: Optional[float] = None,
        timeout: Optional[int] = None,
        halt_on_error: Optional[bool] = True,
    ) -> dict:
        """Poll Batch status to ThreatConnect API.

        .. code-block:: javascript

            {
                "status": "Success",
                "data": {
                    "batchStatus": {
                        "id":3505,
                        "status":"Completed",
                        "errorCount":0,
                        "successCount":0,
                        "unprocessCount":0
                    }
                }
            }

        Args:
            batch_id: The ID returned from the ThreatConnect API for the current batch job.
            retry_seconds: The base number of seconds used for retries when job is not completed.
            back_off: A multiplier to use for backing off on
                each poll attempt when job has not completed.
            timeout: The number of seconds before the poll should timeout.
            halt_on_error: If True any exception will raise an error.

        Returns:
            dict: The batch status returned from the ThreatConnect API.
        """
        # check global setting for override
        if self.halt_on_poll_error is not None:
            halt_on_error = self.halt_on_poll_error

        # initial poll interval
        if self._poll_interval is None and self._batch_data_count is not None:
            # calculate poll_interval base off the number of entries in the batch data
            # with a minimum value of 5 seconds.
            self._poll_interval = max(math.ceil(self._batch_data_count / 300), 5)
        elif self._poll_interval is None:
            # if not able to calculate poll_interval default to 15 seconds
            self._poll_interval = 15

        # poll retry back_off factor
        poll_interval_back_off = float(2.5 if back_off is None else back_off)

        # poll retry seconds
        poll_retry_seconds = int(5 if retry_seconds is None else retry_seconds)

        # poll timeout
        if timeout is None:
            timeout = self.poll_timeout
        else:
            timeout = int(timeout)
        params = {'includeAdditional': 'true'}

        poll_count = 0
        poll_time_total = 0
        data = {}
        while True:
            poll_count += 1
            poll_time_total += self._poll_interval
            time.sleep(self._poll_interval)
            self.tcex.log.info(f'feature=batch, event=progress, poll-time={poll_time_total}')
            try:
                # retrieve job status
                r = self.tcex.session.get(f'/v2/batch/{batch_id}', params=params)
                if not r.ok or 'application/json' not in r.headers.get('content-type', ''):
                    self.tcex.handle_error(545, [r.status_code, r.text], halt_on_error)
                    return data
                data = r.json()
                if data.get('status') != 'Success':
                    self.tcex.handle_error(545, [r.status_code, r.text], halt_on_error)
            except Exception as e:
                self.tcex.handle_error(540, [e], halt_on_error)

            if data.get('data', {}).get('batchStatus', {}).get('status') == 'Completed':
                # store last 5 poll times to use in calculating average poll time
                modifier = poll_time_total * 0.7
                self._poll_interval_times = self._poll_interval_times[-4:] + [modifier]

                weights = [1]
                poll_interval_time_weighted_sum = 0
                for poll_interval_time in self._poll_interval_times:
                    poll_interval_time_weighted_sum += poll_interval_time * weights[-1]
                    # weights will be [1, 1.5, 2.25, 3.375, 5.0625] for all 5 poll times depending
                    # on how many poll times are available.
                    weights.append(weights[-1] * 1.5)

                # pop off the last weight so its not added in to the sum
                weights.pop()

                # calculate the weighted average of the last 5 poll times
                self._poll_interval = math.floor(poll_interval_time_weighted_sum / sum(weights))

                if poll_count == 1:
                    # if completed on first poll, reduce poll interval.
                    self._poll_interval = self._poll_interval * 0.85

                self.tcex.log.debug(f'feature=batch, poll-time={poll_time_total}, status={data}')
                return data

            # update poll_interval for retry with max poll time of 20 seconds
            self._poll_interval = min(
                poll_retry_seconds + int(poll_count * poll_interval_back_off), 20
            )

            # time out poll to prevent App running indefinitely
            if poll_time_total >= timeout:
                self.tcex.handle_error(550, [timeout], True)

    @property
    def poll_timeout(self) -> int:
        """Return current poll timeout value."""
        return self._poll_timeout

    @poll_timeout.setter
    def poll_timeout(self, seconds: int):
        """Set the poll timeout value."""
        self._poll_timeout = int(seconds)

    def process_all(self, process_files: Optional[bool] = True) -> None:
        """Process Batch request to ThreatConnect API.

        Args:
            process_files: Send any document or report attachments to the API.
        """
        while True:
            content = self.data
            file_data = content.pop('file', {})
            if not content.get('group') and not content.get('indicator'):
                break

            # special code for debugging App using batchV2.
            self.write_batch_json(content)

            # store the length of the batch data to use for poll interval calculations
            self.tcex.log.info(
                '''feature=batch, event=process-all, type=group, '''
                f'''count={len(content.get('group')):,}'''
            )
            self.tcex.log.info(
                '''feature=batch, event=process-all, type=indicator, '''
                f'''count={len(content.get('indicator')):,}'''
            )

        if process_files:
            self.process_files(file_data)

    def process_files(self, file_data: dict) -> None:
        """Process Files for Documents and Reports to ThreatConnect API.

        Args:
            file_data: The file data to be processed.
        """
        for xid, content_data in list(file_data.items()):
            del file_data[xid]  # win or loose remove the entry

            # define the saved filename
            api_branch = 'documents' if content_data.get('type') == 'Report' else 'reports'
            fqfn = os.path.join(
                self.debug_path_files,
                f'''{api_branch}--{xid}--{content_data.get('fileName').replace('/', ':')}''',
            )

            # used for debug/testing to prevent upload of previously uploaded file
            if self.debug and xid in self.saved_xids:
                self.tcex.log.debug(
                    f'feature=batch-submit-files, action=skip-previously-saved-file, xid={xid}'
                )
                continue

            if os.path.isfile(fqfn):
                self.tcex.log.debug(
                    f'feature=batch-submit-files, action=skip-previously-saved-file, xid={xid}'
                )
                continue

            # process the file content
            content = content_data.get('fileContent')
            if callable(content):
                content_callable_name = getattr(content, '__name__', repr(content))
                self.tcex.log.trace(
                    f'feature=batch-submit-files, method={content_callable_name}, xid={xid}'
                )
                content = content_data.get('fileContent')(xid)

            if content is None:
                self.tcex.log.warning(f'feature=batch-submit-files, xid={xid}, event=content-null')
                continue

            # write the file to disk
            with open(fqfn, 'wb') as fh:
                fh.write(content)

    def registry_key(
        self, key_name: str, value_name: str, value_type: str, **kwargs
    ) -> RegistryKey:
        """Add Registry Key data to Batch object.

        Args:
            key_name: The key_name value for this Indicator.
            value_name: The value_name value for this Indicator.
            value_type: The value_type value for this Indicator.
            confidence (str, kwargs): The threat confidence for this Indicator.
            date_added (str, kwargs): The date timestamp the Indicator was created.
            last_modified (str, kwargs): The date timestamp the Indicator was last modified.
            rating (str, kwargs): The threat rating for this Indicator.
            xid (str, kwargs): The external id for this Indicator.
            store: (bool, kwargs): Advanced - Defaults to True. If True
                the indicator data will be stored in instance list.

        Returns:
            RegistryKey: An instance of the Registry Key class.
        """
        indicator_obj = RegistryKey(key_name, value_name, value_type, **kwargs)
        return self._indicator(indicator_obj, kwargs.get('store', True))

    def report(self, name: str, **kwargs) -> Report:
        """Add Report data to Batch object.

        Args:
            name: The name for this Group.
            file_name (str): The name for the attached file for this Group.
            date_added (str, kwargs): The date timestamp the Indicator was created.
            file_content (str;method, kwargs): The file contents or callback method to retrieve
                file content.
            publish_date (str, kwargs): The publish datetime expression for this Group.
            xid (str, kwargs): The external id for this Group.
            store: (bool, kwargs): Advanced - Defaults to True. If True
                the indicator data will be stored in instance list.

        Returns:
            Report: An instance of the Report class.
        """
        group_obj = Report(name, **kwargs)
        return self._group(group_obj, kwargs.get('store', True))

    def save(self, resource: Union[dict, object]) -> None:
        """Save group|indicator dict or object to shelve.

        Best effort to save group/indicator data to disk.  If for any reason the save fails
        the data will still be accessible from list in memory.

        Args:
            resource: The Group or Indicator dict or object.
        """
        resource_type = None
        xid = None
        if isinstance(resource, dict):
            resource_type = resource.get('type')
            xid = resource.get('xid')
        else:
            resource_type = resource.type
            xid = resource.xid

        if resource_type is not None and xid is not None:
            saved = True

            if resource_type in self.tcex.group_types:
                try:
                    # groups
                    self.groups_shelf[xid] = resource
                except Exception:
                    saved = False

                if saved:
                    try:
                        del self._groups[xid]
                    except KeyError:
                        # if group was saved twice it would already be delete
                        pass
            elif resource_type in self.tcex.indicator_types_data.keys():
                try:
                    # indicators
                    self.indicators_shelf[xid] = resource
                except Exception:
                    saved = False

                if saved:
                    try:
                        del self._indicators[xid]
                    except KeyError:
                        # if indicator was saved twice it would already be delete
                        pass

    @property
    def saved_groups(self) -> bool:
        """Return True if saved group files exits, else False."""
        if self._saved_groups is None:
            self._saved_groups = False
            if (
                self.enable_saved_file
                and os.path.isfile(self.debug_path_group_shelf)
                and os.access(self.debug_path_group_shelf, os.R_OK)
            ):
                self._saved_groups = True
                self.tcex.log.debug('feature=batch, event=saved-groups-file-found')
        return self._saved_groups

    @property
    def saved_indicators(self) -> bool:
        """Return True if saved indicators files exits, else False."""
        if self._saved_indicators is None:
            self._saved_indicators = False
            if (
                self.enable_saved_file
                and os.path.isfile(self.debug_path_indicator_shelf)
                and os.access(self.debug_path_indicator_shelf, os.R_OK)
            ):
                self._saved_indicators = True
                self.tcex.log.debug('feature=batch, event=saved-indicator-file-found')
        return self._saved_indicators

    @property
    def saved_xids(self) -> list:
        """Return previously saved xids."""
        if self._saved_xids is None:
            self._saved_xids = []
            if self.debug:
                if os.path.isfile(self.debug_path_xids) and os.access(
                    self.debug_path_xids, os.R_OK
                ):
                    with open(self.debug_path_xids) as fh:
                        self._saved_xids = fh.read().splitlines()
        return self._saved_xids

    @saved_xids.setter
    def saved_xids(self, xid: str):
        """Append xid to xids saved file."""
        with open(self.debug_path_xids, 'a') as fh:
            fh.write(f'{xid}\n')

    @property
    def settings(self) -> dict:
        """Return batch job settings."""
        _settings = {
            'action': self._action,
            # not supported in v2 batch
            # 'attributeWriteType': self._attribute_write_type,
            'attributeWriteType': 'Replace',
            'haltOnError': str(self._halt_on_error).lower(),
            'owner': self._owner,
            'version': 'V2',
        }
        if self._playbook_triggers_enabled is not None:
            _settings['playbookTriggersEnabled'] = str(self._playbook_triggers_enabled).lower()
        if self._hash_collision_mode is not None:
            _settings['hashCollisionMode'] = self._hash_collision_mode
        if self._file_merge_mode is not None:
            _settings['fileMergeMode'] = self._file_merge_mode
        return _settings

    def signature(
        self, name: str, file_name: str, file_type: str, file_text: str, **kwargs
    ) -> Signature:
        """Add Signature data to Batch object.

        Valid file_types:
        + Snort ®
        + Suricata
        + YARA
        + ClamAV ®
        + OpenIOC
        + CybOX ™
        + Bro
        + Regex
        + SPL - Splunk ® Search Processing Language

        Args:
            name: The name for this Group.
            file_name: The name for the attached signature for this Group.
            file_type: The signature type for this Group.
            file_text: The signature content for this Group.
            date_added (str, kwargs): The date timestamp the Indicator was created.
            xid (str, kwargs): The external id for this Group.
            store: (bool, kwargs): Advanced - Defaults to True. If True
                the indicator data will be stored in instance list.

        Returns:
            Signature: An instance of the Signature class.
        """
        group_obj = Signature(name, file_name, file_type, file_text, **kwargs)
        return self._group(group_obj, kwargs.get('store', True))

    def submit(
        self,
        poll: Optional[bool] = True,
        errors: Optional[bool] = True,
        process_files: Optional[bool] = True,
        halt_on_error: Optional[bool] = True,
    ) -> dict:
        """Submit Batch request to ThreatConnect API.

        By default this method will submit the job request and data and if the size of the data
        is below the value **synchronousBatchSaveLimit** set in System Setting it will process
        the request synchronously and return the batch status.  If the size of the batch is greater
        than the value set the batch job will be queued.
        Errors are not retrieve automatically and need to be enabled.

        If any of the submit, poll, or error methods fail the entire submit will halt at the point
        of failure. The behavior can be changed by setting halt_on_error to False.

        Each of these methods can also be called on their own for greater control of the submit
        process.

        Args:
            poll: If True poll batch for status.
            errors: If True retrieve any batch errors (only if poll is True).
            process_files: If true send any document or report attachments to the API.
            halt_on_error: If True any exception will raise an error.

        Returns.
            dict: The Batch Status from the ThreatConnect API.
        """
        # get file, group, and indicator data
        content = self.data

        # pop any file content to pass to submit_files
        file_data = content.pop('file', {})
        batch_data = (
            self.submit_create_and_upload(content=content, halt_on_error=halt_on_error)
            .get('data', {})
            .get('batchStatus', {})
        )
        batch_id = batch_data.get('id')
        if batch_id is not None:
            self.tcex.log.info(f'feature=batch, event=submit, batch-id={batch_id}')
            # job hit queue
            if poll:
                # poll for status
                batch_data = (
                    self.poll(batch_id=batch_id, halt_on_error=halt_on_error)
                    .get('data', {})
                    .get('batchStatus')
                )
                if errors:
                    # retrieve errors
                    error_groups = batch_data.get('errorGroupCount', 0)
                    error_indicators = batch_data.get('errorIndicatorCount', 0)
                    if error_groups > 0 or error_indicators > 0:
                        batch_data['errors'] = self.errors(batch_id)
            else:
                # can't process files if status is unknown (polling must be enabled)
                process_files = False

        if process_files:
            # submit file data after batch job is complete
            self._file_threads.append(
                self.submit_thread(
                    name='submit-files', target=self.submit_files, args=(file_data, halt_on_error,),
                )
            )
        return batch_data

    def submit_all(
        self,
        poll: Optional[bool] = True,
        errors: Optional[bool] = True,
        process_files: Optional[bool] = True,
        halt_on_error: Optional[bool] = True,
    ) -> dict:
        """Submit Batch request to ThreatConnect API.

        By default this method will submit the job request and data and if the size of the data
        is below the value **synchronousBatchSaveLimit** set in System Setting it will process
        the request synchronously and return the batch status.  If the size of the batch is greater
        than the value set the batch job will be queued.
        Errors are not retrieve automatically and need to be enabled.

        If any of the submit, poll, or error methods fail the entire submit will halt at the point
        of failure. The behavior can be changed by setting halt_on_error to False.

        Each of these methods can also be called on their own for greater control of the submit
        process.

        Args:
            poll: If True poll batch for status.
            errors: If True retrieve any batch errors (only if poll is True).
            process_files: If true send any document or report attachments to the API.
            halt_on_error: If True any exception will raise an error.

        Returns.
            dict: The Batch Status from the ThreatConnect API.
        """
        batch_data_array = []
        file_data = {}
        while True:
            batch_data = {}
            batch_id = None

            # get file, group, and indicator data
            content = self.data

            # break loop when end of data is reached
            if not content.get('group') and not content.get('indicator'):
                break

            if self.action.lower() == 'delete':
                # no need to process files on a delete batch job
                process_files = False

                # while waiting of FR for delete support in createAndUpload submit delete request
                # the old way (submit job + submit data), still using V2.
                if len(self) > 0:  # pylint: disable=len-as-condition
                    batch_id = self.submit_job(halt_on_error)
                    if batch_id is not None:
                        batch_data = self.submit_data(
                            batch_id=batch_id, content=content, halt_on_error=halt_on_error
                        )
                else:
                    batch_data = {}
            else:
                # pop any file content to pass to submit_files
                file_data = content.pop('file', {})
                batch_data = (
                    self.submit_create_and_upload(content=content, halt_on_error=halt_on_error)
                    .get('data', {})
                    .get('batchStatus', {})
                )
                batch_id = batch_data.get('id')

            if batch_id is not None:
                self.tcex.log.info(f'feature=batch, event=status, batch-id={batch_id}')
                # job hit queue
                if poll:
                    # poll for status
                    batch_data = (
                        self.poll(batch_id, halt_on_error=halt_on_error)
                        .get('data', {})
                        .get('batchStatus')
                    )
                    if errors:
                        # retrieve errors
                        error_count = batch_data.get('errorCount', 0)
                        error_groups = batch_data.get('errorGroupCount', 0)
                        error_indicators = batch_data.get('errorIndicatorCount', 0)
                        if error_count > 0 or error_groups > 0 or error_indicators > 0:
                            batch_data['errors'] = self.errors(batch_id)
                else:
                    # can't process files if status is unknown (polling must be enabled)
                    process_files = False

            if process_files:
                # submit file data after batch job is complete
                self._file_threads.append(
                    self.submit_thread(
                        name='submit-files',
                        target=self.submit_files,
                        args=(file_data, halt_on_error,),
                    )
                )
            batch_data_array.append(batch_data)

            # write errors for debugging
            self.write_error_json(batch_data.get('errors'))

        return batch_data_array

    def submit_callback(
        self,
        callback: Callable[..., Any],
        content: Optional[dict] = None,
        halt_on_error: Optional[bool] = True,
    ) -> bool:
        """Submit batch data to ThreatConnect and poll in a separate thread.

        The "normal" submit methods run in serial which will block when the batch poll is running.
        Using this method the submit is done in serial, but the poll method is run in a thread,
        which should allow the App to continue downloading and processing data while the batch
        poll process is running. Only one batch submission is allowed at a time so that any
        critical errors returned from batch can be handled before submitting a new batch job.

        Args:
            callback: The callback method that will handle
                the batch status when polling is complete.
            content: The dict of groups and indicator data (e.g., {"group": [], "indiciator": []}).
            halt_on_error: If True the process should halt if any errors are encountered.

        Raises:
            RuntimeError: Raised on invalid callback method.

        Returns:
            bool: False when there is not data to process, else True
        """
        # user provided content or grab content from local group/indicator lists
        if content is not None:
            # process content
            pass
        else:
            content = self.data
        file_data = content.pop('file', {})

        # return False when end of data is reached
        if not content.get('group') and not content.get('indicator'):
            return False

        # block here is there is already a batch submission being processed
        if hasattr(self._submit_thread, 'is_alive'):
            self.tcex.log.info(
                'feature=batch, event=progress, status=blocked, '
                f'is-alive={self._submit_thread.is_alive()}'
            )
            self._submit_thread.join()
            self.tcex.log.debug(
                'feature=batch, event=progress, status=released, '
                f'is-alive={self._submit_thread.is_alive()}'
            )

        # submit the data and collect the response
        batch_data: dict = (
            self.submit_create_and_upload(content=content, halt_on_error=halt_on_error)
            .get('data', {})
            .get('batchStatus', {})
        )
        self.tcex.log.trace(f'feature=batch, event=submit-callback, batch-data={batch_data}')

        # launch batch polling in a thread
        self._submit_thread = self.submit_thread(
            name='submit-poll',
            target=self.submit_callback_thread,
            args=(batch_data, callback, file_data),
        )

        return True

    def submit_callback_thread(
        self,
        batch_data: int,
        callback: Callable[..., Any],
        file_data: dict,
        halt_on_error: Optional[bool] = True,
    ) -> None:
        """Submit data in a thread."""
        batch_id = batch_data.get('id')
        self.tcex.log.info(f'feature=batch, event=progress, batch-id={batch_id}')
        if batch_id:
            # when batch_id is None it indicates that batch submission was small enough to be
            # processed inline (without being queued)

            # poll for status
            batch_status = (
                self.poll(batch_id, halt_on_error=halt_on_error).get('data', {}).get('batchStatus')
            )

            # retrieve errors
            error_count = batch_status.get('errorCount', 0)
            error_groups = batch_status.get('errorGroupCount', 0)
            error_indicators = batch_status.get('errorIndicatorCount', 0)
            if error_count > 0 or error_groups > 0 or error_indicators > 0:
                batch_status['errors'] = self.errors(batch_id)
        else:
            batch_status = batch_data

        # launch file upload in a thread *after* batch status is returned. while only one batch
        # submission thread is allowed, there is no limit on file upload threads. the upload
        # status returned by file upload will be ignored when running in a thread.
        if file_data:
            self._file_threads.append(
                self.submit_thread(
                    name='submit-files', target=self.submit_files, args=(file_data, halt_on_error,),
                )
            )

        # send batch_status to callback
        if callable(callback):
            self.tcex.log.debug('feature=batch, event=calling-callback')
            try:
                callback(batch_status)
            except Exception as e:
                self.tcex.log.warning(f'feature=batch, event=callback-error, err="""{e}"""')

    def submit_create_and_upload(self, content: dict, halt_on_error: Optional[bool] = True) -> dict:
        """Submit Batch request to ThreatConnect API.

        Args:
            content: The dict of groups and indicator data.
            halt_on_error: If True the process should halt if any errors are encountered.

        Returns.
            dict: The Batch Status from the ThreatConnect API.
        """
        # check global setting for override
        if self.halt_on_batch_error is not None:
            halt_on_error = self.halt_on_batch_error

        # special code for debugging App using batchV2.
        self.write_batch_json(content)

        # store the length of the batch data to use for poll interval calculations
        self.tcex.log.info(
            '''feature=batch, event=submit-create-and-upload, type=group, '''
            f'''count={len(content.get('group')):,}'''
        )
        self.tcex.log.info(
            '''feature=batch, event=submit-create-and-upload, type=indicator, '''
            f'''count={len(content.get('indicator')):,}'''
        )

        try:
            files = (('config', json.dumps(self.settings)), ('content', json.dumps(content)))
            params = {'includeAdditional': 'true'}
            r = self.tcex.session.post('/v2/batch/createAndUpload', files=files, params=params)
            if not r.ok or 'application/json' not in r.headers.get('content-type', ''):
                self.tcex.handle_error(10510, [r.status_code, r.text], halt_on_error)
            return r.json()
        except Exception as e:
            self.tcex.handle_error(10505, [e], halt_on_error)

        return {}

    def submit_data(
        self, batch_id: int, content: dict, halt_on_error: Optional[bool] = True
    ) -> dict:
        """Submit Batch request to ThreatConnect API.

        Args:
            batch_id: The batch id of the current job.
            content: The dict of groups and indicator data.
            halt_on_error (Optional[bool] = True): If True the process
                should halt if any errors are encountered.

        Returns:
            dict: The response data
        """
        # check global setting for override
        if self.halt_on_batch_error is not None:
            halt_on_error = self.halt_on_batch_error

        # store the length of the batch data to use for poll interval calculations
        self._batch_data_count = len(content.get('group')) + len(content.get('indicator'))
        self.tcex.log.info(
            f'feature=batch, action=submit-data, batch-size={self._batch_data_count:,}'
        )

        headers = {'Content-Type': 'application/octet-stream'}
        try:
            r = self.tcex.session.post(f'/v2/batch/{batch_id}', headers=headers, json=content)
            if not r.ok or 'application/json' not in r.headers.get('content-type', ''):
                self.tcex.handle_error(10525, [r.status_code, r.text], halt_on_error)
            return r.json()
        except Exception as e:
            self.tcex.handle_error(10520, [e], halt_on_error)

        return None

    def submit_files(self, file_data: dict, halt_on_error: Optional[bool] = True) -> dict:
        """Submit Files for Documents and Reports to ThreatConnect API.

        Critical Errors

        * There is insufficient document storage allocated to this account.

        Args:
            halt_on_error: If True any exception will raise an error.
            file_data: The file data to be submitted.

        Returns:
            dict: The upload status for each xid.
        """
        # check global setting for override
        if self.halt_on_file_error is not None:
            halt_on_error = self.halt_on_file_error

        upload_status = []
        self.tcex.log.info(f'feature=batch, action=submit-files, count={len(file_data)}')
        for xid, content_data in list(file_data.items()):
            del file_data[xid]  # win or loose remove the entry
            status = True

            # used for debug/testing to prevent upload of previously uploaded file
            if self.debug and xid in self.saved_xids:
                self.tcex.log.debug(
                    f'feature=batch-submit-files, action=skip-previously-saved-file, xid={xid}'
                )
                continue

            # process the file content
            content = content_data.get('fileContent')
            if callable(content):
                try:
                    content_callable_name = getattr(content, '__name__', repr(content))
                    self.tcex.log.trace(
                        f'feature=batch-submit-files, method={content_callable_name}, xid={xid}'
                    )
                    content = content_data.get('fileContent')(xid)
                except Exception as e:
                    self.tcex.log.warning(
                        f'feature=batch, event=file-download-exception, err="""{e}"""'
                    )

            if content is None:
                upload_status.append({'uploaded': False, 'xid': xid})
                self.tcex.log.warning(f'feature=batch-submit-files, xid={xid}, event=content-null')
                continue

            api_branch = 'documents'
            if content_data.get('type') == 'Report':
                api_branch = 'reports'

            if self.debug and content_data.get('fileName'):
                # special code for debugging App using batchV2.
                fqfn = os.path.join(
                    self.debug_path_files,
                    f'''{api_branch}--{xid}--{content_data.get('fileName').replace('/', ':')}''',
                )
                with open(fqfn, 'wb') as fh:
                    if not isinstance(content, bytes):
                        content = content.encode()
                    fh.write(content)

            # Post File
            url = f'/v2/groups/{api_branch}/{xid}/upload'
            headers = {'Content-Type': 'application/octet-stream'}
            params = {'owner': self._owner, 'updateIfExists': 'true'}
            r = self.submit_file_content('POST', url, content, headers, params, halt_on_error)
            if r.status_code == 401:
                # use PUT method if file already exists
                self.tcex.log.info('feature=batch, event=401-from-post, action=switch-to-put')
                r = self.submit_file_content('PUT', url, content, headers, params, halt_on_error)
            if not r.ok:
                status = False
                self.tcex.handle_error(585, [r.status_code, r.text], halt_on_error)
            elif self.debug and self.enable_saved_file and xid not in self.saved_xids:
                # save xid "if" successfully uploaded and not already saved
                self.saved_xids = xid

            self.tcex.log.info(
                f'feature=batch, event=file-upload, status={r.status_code}, '
                f'xid={xid}, remaining={len(file_data)}'
            )
            upload_status.append({'uploaded': status, 'xid': xid})

        return upload_status

    def submit_file_content(
        self,
        method: str,
        url: str,
        data: Union[bytes, str],
        headers: dict,
        params: dict,
        halt_on_error: Optional[bool] = True,
    ) -> object:
        """Submit File Content for Documents and Reports to ThreatConnect API.

        Args:
            method: The HTTP method for the request (POST, PUT).
            url: The URL for the request.
            data: The body (data) for the request.
            headers: The headers for the request.
            params: The query string parameters for the request.
            halt_on_error: If True any exception will raise an error.

        Returns:
            requests.models.Response: The response from the request.
        """
        r = None
        try:
            r = self.tcex.session.request(method, url, data=data, headers=headers, params=params)
        except Exception as e:
            self.tcex.handle_error(580, [e], halt_on_error)
        return r

    def submit_job(self, halt_on_error: Optional[bool] = True) -> int:
        """Submit Batch request to ThreatConnect API.

        Args:
            halt_on_error: If True any exception will raise an error.

        Returns:
            int: The batch id from the API response.
        """
        # check global setting for override
        if self.halt_on_batch_error is not None:
            halt_on_error = self.halt_on_batch_error

        try:
            r = self.tcex.session.post('/v2/batch', json=self.settings)
        except Exception as e:
            self.tcex.handle_error(10505, [e], halt_on_error)

        if not r.ok or 'application/json' not in r.headers.get('content-type', ''):
            self.tcex.handle_error(10510, [r.status_code, r.text], halt_on_error)
        data = r.json()
        if data.get('status') != 'Success':
            self.tcex.handle_error(10510, [r.status_code, r.text], halt_on_error)
        self.tcex.log.debug(f'feature=batch, event=submit-job, status={data}')
        return data.get('data', {}).get('batchId')

    def submit_thread(
        self,
        name: str,
        target: Callable[[], bool],
        args: Optional[tuple] = None,
        kwargs: Optional[dict] = None,
    ) -> None:
        """Start a submit thread.

        Args:
            name: The name of the thread.
            target: The method to call for the thread.
            args: The args to pass to the target method.
            kwargs: Additional args.
        """
        self.tcex.log.info(f'feature=batch, event=submit-thread, name={name}')
        args = args or ()
        t = None
        try:
            t = threading.Thread(name=name, target=target, args=args, kwargs=kwargs, daemon=True)
            t.start()
        except Exception:
            self.tcex.log.trace(traceback.format_exc())
        return t

    def threat(self, name: str, **kwargs) -> Threat:
        """Add Threat data to Batch object

        Args:
            name: The name for this Group.
            date_added (str, kwargs): The date timestamp the Indicator was created.
            xid (str, kwargs): The external id for this Group.
            store: (bool, kwargs): Advanced - Defaults to True. If True
                the indicator data will be stored in instance list.

        Returns:
            Threat: An instance of the Threat class.
        """
        group_obj = Threat(name, **kwargs)
        return self._group(group_obj, kwargs.get('store', True))

    def user_agent(self, text: str, **kwargs) -> UserAgent:
        """Add User Agent data to Batch object

        Args:
            text: The value for this Indicator.
            confidence (str, kwargs): The threat confidence for this Indicator.
            date_added (str, kwargs): The date timestamp the Indicator was created.
            last_modified (str, kwargs): The date timestamp the Indicator was last modified.
            rating (str, kwargs): The threat rating for this Indicator.
            xid (str, kwargs): The external id for this Indicator.
            store: (bool, kwargs): Advanced - Defaults to True. If True
                the indicator data will be stored in instance list.

        Returns:
            UserAgent: An instance of the UserAgent class.
        """
        indicator_obj = UserAgent(text, **kwargs)
        return self._indicator(indicator_obj, kwargs.get('store', True))

    def url(self, text: str, **kwargs) -> URL:
        """Add URL Address data to Batch object.

        Args:
            text (str): The value for this Indicator.
            confidence (str, kwargs): The threat confidence for this Indicator.
            date_added (str, kwargs): The date timestamp the Indicator was created.
            last_modified (str, kwargs): The date timestamp the Indicator was last modified.
            rating (str, kwargs): The threat rating for this Indicator.
            xid (str, kwargs): The external id for this Indicator.
            store: (bool, kwargs): Advanced - Defaults to True. If True
                the indicator data will be stored in instance list.

        Returns:
            URL: An instance of the URL class.
        """
        indicator_obj = URL(text, **kwargs)
        return self._indicator(indicator_obj, kwargs.get('store', True))

    def write_error_json(self, errors: list) -> None:
        """Write the errors to a JSON file for debuging purposes.

        Args:
            errors: A list of errors to write out.
        """
        if self.debug:
            if not errors:
                errors = []
            # get timestamp as a string without decimal place and consistent length
            timestamp = str(int(time.time() * 10000000))
            error_json_file = os.path.join(self.debug_path_batch, f'errors-{timestamp}.json.gz')
            with gzip.open(error_json_file, mode='wt', encoding='utf-8') as fh:
                json.dump(errors, fh)

    def write_batch_json(self, content: dict) -> None:
        """Write batch json data to a file."""
        if self.debug and content:
            # get timestamp as a string without decimal place and consistent length
            timestamp = str(int(time.time() * 10000000))
            batch_json_file = os.path.join(self.debug_path_batch, f'batch-{timestamp}.json.gz')
            with gzip.open(batch_json_file, mode='wt', encoding='utf-8') as fh:
                json.dump(content, fh)

    @property
    def group_len(self) -> int:
        """Return the number of current groups."""
        return len(self.groups) + len(self.groups_shelf)

    @property
    def indicator_len(self) -> int:
        """Return the number of current indicators."""
        return len(self.indicators) + len(self.indicators_shelf)

    def __len__(self) -> int:
        """Return the number of groups and indicators."""
        return self.group_len + self.indicator_len

    def __str__(self) -> str:  # pragma: no cover
        """Return string represtentation of object."""
        groups = []
        for group_data in self.groups.values():
            if isinstance(group_data, dict):
                groups.append(group_data)
            else:
                groups.append(group_data.data)
        for group_data in self.groups_shelf.values():
            if isinstance(group_data, dict):
                groups.append(group_data)
            else:
                groups.append(group_data.data)

        indicators = []
        for indicator_data in self.indicators.values():
            if isinstance(indicator_data, dict):
                indicators.append(indicator_data)
            else:
                indicators.append(indicator_data.data)
        for indicator_data in self.indicators_shelf.values():
            if isinstance(indicator_data, dict):
                indicators.append(indicator_data)
            else:
                indicators.append(indicator_data.data)

        data = {'group': groups, 'indicators': indicators}
        return json.dumps(data, indent=4, sort_keys=True)



"""Key Value Store module for TcEx Framework"""
from .key_value_api import KeyValueApi  # noqa: F401
from .key_value_redis import KeyValueRedis  # noqa: F401
from .redis_client import RedisClient  # noqa: F401



"""ThreatConnect TI Campaign"""
from ..group import Group


class Campaign(Group):
    """Unique API calls for Campaign API Endpoints

    Args:
        tcex (TcEx): An instantiated instance of TcEx object.
        name (str, kwargs): [Required for Create] The name for this Group.
        owner (str, kwargs): The name for this Group. Default to default Org when not provided
        first_seen (str, kwargs): The first seen datetime expression for this Group.
    """

    def __init__(self, tcex, **kwargs):
        """Initialize Class Properties."""
        super().__init__(
            tcex, sub_type='Campaign', api_entity='campaign', api_branch='campaigns', **kwargs
        )

    def first_seen(self, first_seen):
        """Update the campaign with the new first_seen date.

        Args:
            first_seen (str): The first_seen date. Converted to %Y-%m-%dT%H:%M:%SZ date format

        Returns:
            requests.Response: The response from the API call.
        """
        if not self.can_update():
            self._tcex.handle_error(910, [self.type])

        first_seen = self._utils.datetime.format_datetime(
            first_seen, date_format='%Y-%m-%dT%H:%M:%SZ'
        )
        self._data['firstSeen'] = first_seen
        request = {'firstSeen': first_seen}
        return self.tc_requests.update(self.api_type, self.api_branch, self.unique_id, request)



#!/usr/bin/env python
"""TcEx App Init."""
from ..app_config_object.templates import DownloadTemplates
from .bin import Bin


class Init(Bin):
    """Install required modules for ThreatConnect Job or Playbook App.

    Args:
        _args (namespace): The argparser args Namespace.
    """

    def __init__(self, _args):
        """Initialize Class properties."""
        super().__init__(_args)
        self.download_template = DownloadTemplates(branch=self.args.branch)



"""not_in argument validator."""
# standard library
from typing import Any, Callable, List

from .validation_exception import ValidationError


def not_in(invalid_values: List[Any]) -> Callable[..., None]:
    """Validate a String is in not equal to any of the values in invalid_values.

    Allowed argument types: String, StringArray, TCEntity, TCEntityArray


    Args:
        invalid_values: List of values that the argument value cannot be.
        **kwargs ():
            strip (bool): call .strip() on the argument value

    Returns:
        A validator function that can be used in the validators argument to @ReadArg.
    """

    def _validate(value: str, arg_name: str, label: str) -> None:
        """Run validation on input data.

        Args:
            value: The input data to validate.
            arg_name: The name of the input arg.
            label: The label displayed to the user.

        Raises:
            ValidationError: raised on validation failure.
        """
        if value in invalid_values:
            formatted_values = ', '.join(
                [f'"{n}"' if isinstance(n, str) else str(n) for n in invalid_values]
            )

            raise ValidationError(f'"{label}" ({arg_name}) cannot be in [{formatted_values}]')

    return _validate



"""Parser for STIX Domain Name Object.

see: https://docs.oasis-open.org/cti/stix/v2.1/csprd01/stix-v2.1-csprd01.html#_Toc16070687
"""
# standard library
from typing import Iterable, Union

# third-party
from stix2 import DomainName

# first-party
from tcex.stix.model import StixModel


class StixDomainNameObject(StixModel):
    """Parser for STIX Domain Name Object.

    see: https://docs.oasis-open.org/cti/stix/v2.1/csprd01/stix-v2.1-csprd01.html#_Toc16070687
    """

    def produce(self, tc_data: Union[list, dict], **kwargs) -> Iterable[DomainName]:
        """Produce STIX 2.0 JSON object from TC API response."""
        if isinstance(tc_data, list) and len(tc_data) > 0 and 'summary' in tc_data[0]:
            indicator_field = 'summary'
        else:
            indicator_field = 'hostName'

        parse_map = {
            'type': 'domain-name',
            'spec_version': '2.1',
            'id': '@.id',
            'value': f'@."{indicator_field}"',
        }

        yield from (DomainName(**stix_data) for stix_data in self._map(tc_data, parse_map))

    def consume(self, stix_data: Union[list, dict]):
        """Produce a ThreatConnect object from a STIX 2.0 JSON object."""
        yield from self._map(
            stix_data,
            {
                'type': 'Host',
                'hostName': '@.value',
                'xid': '@.id',
                'attributes': [{'type': 'External ID', 'value': '@.id'}],
            },
        )



"""Parsers for STIX Cyber Observables."""
# flake8: noqa
from .autonomous_system import StixASObject
from .domain_name import StixDomainNameObject
from .email_address import StixEmailAddressObject
from .ip_addr import StixIPv4Object, StixIPv6Object
from .registry_key import StixRegistryKeyObject
from .url import StixURLObject



"""TcEx Framework Module for working with Cache in the ThreatConnect Platform."""
# standard library
from datetime import datetime, timedelta
from typing import Callable, Optional


class Cache:
    """TcEx Cache Class.

    Args:
        tcex (TcEx): An instance of TcEx.
        domain (str): A value of "organization" or "local".
        data_type (str): A free form type name for the data.
        ttl_seconds (Optional[int] = None): Number of seconds the cache is valid.
        mapping (Optional[dict] = None): ElasticSearch mappings data.
    """

    def __init__(
        self,
        tcex: object,
        domain: str,
        data_type: str,
        ttl_seconds: Optional[int] = None,
        mapping: Optional[dict] = None,
    ):
        """Initialize class properties."""
        self.tcex: object = tcex

        # properties
        self.ttl_seconds: Optional[int] = ttl_seconds
        self.ds: 'tcex.datastore' = self.tcex.datastore(domain, data_type, mapping)

        # Warranty void if any of these are changed.  Don't touch.
        self._cache_data_key: str = 'cache-data'
        self._cache_date_key: str = 'cache-date'

    def add(self, rid: str, data: dict, raise_on_error: Optional[bool] = True) -> dict:
        """Write cache data to the data store.

        **Example Response**

        .. code-block:: json

            {
              "_index": "$local.usr5_pytest",
              "_type": "pytest",
              "_id": "cache-one",
              "_version": 10,
              "result": "updated",
              "_shards": {
                  "total": 2,
                  "successful": 1,
                  "failed": 0
              },
              "_seq_no": 10,
              "_primary_term": 1
            }

        Args:
            rid (str): The record identifier.
            data (dict): The record data.
            raise_on_error (Optional[bool] = True): If True and not r.ok this
                method will raise a RunTimeError.

        Returns:
            dict : The response dict
        """
        data: dict = {
            self._cache_date_key: datetime.utcnow().isoformat(),
            self._cache_data_key: data,
        }
        return self.ds.post(rid, data, raise_on_error)

    def delete(self, rid: str, raise_on_error: Optional[bool] = True) -> dict:
        """Write cache data to the data store.

        **Example Response**

        .. code-block:: json

            {
              "_index": "$local.usr5_pytest",
              "_type": "pytest",
              "_id": "cache-delete",
              "_version": 2,
              "result": "deleted",
              "_shards": {
                "total": 2,
                "successful": 1,
                "failed": 0
              },
              "_seq_no": 34,
              "_primary_term": 1
            }

        Args:
            rid (str): The record identifier.
            raise_on_error (Optional[bool] = True): If True and not r.ok this
                 method will raise a RunTimeError.

        Returns:
            dict : The response dict.
        """
        return self.ds.delete(rid, raise_on_error)

    def get(
        self,
        rid: str,
        data_callback: Optional[Callable[[str], dict]] = None,
        raise_on_error: Optional[bool] = True,
    ) -> dict:
        """Get cached data from the data store.

        **Example Response**

        .. code-block:: json

            {
              "cache-date": "2020-07-31T11:44:53.851116",
              "cache-data": {
                "results": "cached"
              }
            }

        Args:
            rid (str): The record identifier.
            data_callback (Optional[Callable] = None): A method that will return the data.
            raise_on_error (Optional[bool] = True): If True and not r.ok this
                method will raise a RunTimeError.

        Returns:
            dict : The cached data.
        """
        cache_data = None
        ds_data: dict = self.ds.get(rid, raise_on_error=False)

        if ds_data is None:
            # default the response when TC API doesn't return a value
            ds_data = {'found': False}

        if ds_data is not None:
            expired = False
            if ds_data.get('found') is True:
                cache_data: dict = ds_data.get('_source', {})
                cache_date: str = cache_data.get(self._cache_date_key)
                if self._is_cache_expired(cache_date):
                    cache_data = None
                    expired = True
                    self.tcex.log.debug(f'Cached data is expired for ({rid}).')

            if expired or ds_data.get('found') is False:
                # when cache is expired or does not exist use callback to get data if possible
                if callable(data_callback):
                    # cache_data = self._encode_data(data_callback(rid))
                    cache_data: Optional[dict] = data_callback(rid)
                    self.tcex.log.debug(f'Using callback data for ({rid}).')
                    if cache_data:
                        cache_data = self.update(
                            rid, cache_data, raise_on_error
                        )  # update the cache data
            else:
                self.tcex.log.debug(f'Using cached data for ({rid}).')

        return cache_data

    def update(self, rid: str, data: dict, raise_on_error: Optional[bool] = True) -> dict:
        """Write updated cache data to the DataStore.

        **Example Response**

        .. code-block:: json

            {
              "cache-date": "2020-07-31T11:44:53.851116",
              "cache-data": {
                "one": 1
              }
            }

        Args:
            rid (str): The record identifier.
            data (dict): The record data.
            raise_on_error (Optional[bool] = True): If True and not r.ok this
               method will raise a RunTimeError.

        Returns:
            dict : The cached data.
        """
        cache_date = datetime.utcnow().isoformat()
        # cache_data = self._encode_data(data)
        data = {self._cache_date_key: cache_date, self._cache_data_key: data}
        self.ds.put(rid, data, raise_on_error)
        return data

    def _is_cache_expired(self, cached_date: str) -> bool:
        """Return True if the provided cache data is expired.

        Args:
            cached_date (str): The cache date value.

        Returns:
            bool: True if cache data is expired.
        """

        if self.ttl_seconds is None or self.ttl_seconds == 0:
            return True  # if ttl_is 0 or None, all cached data is always invalid.

        cached_datetime = self.tcex.utils.datetime.any_to_datetime(cached_date)
        cache_expires = (cached_datetime + timedelta(seconds=self.ttl_seconds)).timestamp()
        return cache_expires < datetime.utcnow().timestamp()



"""in_range argument validator."""
# standard library
from typing import Any, Callable

from .validation_exception import ValidationError


def in_range(
    min: float, max: float, **kwargs  # pylint: disable=redefined-builtin
) -> Callable[..., None]:
    """Validate a value is in the range min <= value <= max.

    For a StringArray, validates that every element in the array is min <= value <= max.

    Allowed argument types: String, StringArray


    Args:
        min: the minimum valid value for the argument
        max: the maximum valid value for the argument
        **kwargs ():
            allow_none (bool): If none or '' values are ok in a StringArray.

    Returns:
        A validator function that can be used in the validators argument to @ReadArg.
    """

    def _validate(value: Any, arg_name: str, label: str) -> None:
        """Run validation on input data.

        Args:
            value: The input data to validate.
            arg_name: The name of the input arg.
            label: The label displayed to the user.

        Raises:
            ValidationError: raised on validation failure.
        """
        allow_none = kwargs.get('allow_none', False)

        if not isinstance(value, list):
            value = [value]

        for v in value:
            if (v is None or v == '') and allow_none:
                continue

            if not min <= v <= max:
                raise ValidationError(f'"{label}" ({arg_name}) is not between {min} and {max}.')

    return _validate



"""Cache Handler Class"""
# standard library
import logging


class CacheHandler(logging.Handler):
    """Logger handler for caching event until all handlers are availabe."""

    def __init__(self, max_cache=100):
        """Initialize Class properties.

        Args:
            max_cache (int): The maximum numbers of records to cache.
        """
        super().__init__()
        self.max_cache = max_cache
        self._events = []

    def emit(self, record):
        """Emit a record.

        Args:
            record (obj): The record to be logged.
        """
        # cache log events
        if len(self._events) <= self.max_cache:
            self._events.append(record)

    @property
    def events(self):
        """Return a copy and clear self._events."""
        events = list(self._events)
        self._events = []
        return events



"""Stage Data Testing Module"""
# standard library
import base64
import binascii
import json
import os
import sys


# pylint: disable=no-self-use
class Stager:
    """Stage Data class"""

    def __init__(self, tcex, log):
        """Initialize class properties"""
        # self.args = tcex.args  # required for args to be parsed
        self.log = log
        self.tcex = tcex
        # TODO: validate this
        self.tcex.logger.update_handler_level('error')

        # properties
        self._redis = None
        self._threatconnect = None

    @property
    def redis(self):
        """Get the current instance of Redis for staging data"""
        if not self._redis:
            self._redis = Redis(self)
        return self._redis

    @property
    def threatconnect(self):
        """Get the current instance of ThreatConnect for staging data"""
        if not self._threatconnect:
            self._threatconnect = ThreatConnect(self)
        return self._threatconnect


class Redis:
    """Stages the Redis Data"""

    def __init__(self, provider):
        """Initialize class properties."""
        self.provider = provider
        self.redis_client = provider.tcex.redis_client
        self.log = self.provider.log

    def from_dict(self, staging_data):
        """Stage redis data from dict"""
        for variable, data in staging_data.items():
            variable_type = self.provider.tcex.playbook.variable_type(variable)

            self.log.data('stage', 'variable', variable)
            self.log.data('stage', 'data', data)
            if data is not None and variable_type == 'Binary':
                data = self._decode_binary(data, variable)
            elif data is not None and variable_type == 'BinaryArray':
                data = [self._decode_binary(d, variable) for d in data]
            self.provider.tcex.playbook.create(variable, data)

    def stage(self, variable, data):
        """Stage data in redis"""
        self.provider.tcex.playbook.create(variable, data)

    def delete_context(self, context):
        """Delete data in redis"""
        keys = self.redis_client.hkeys(context)
        if keys:
            return self.redis_client.hdel(context, *keys)
        return 0

    @staticmethod
    def _decode_binary(binary_data, variable):
        """Base64 decode binary data."""
        try:
            data = None
            if binary_data is not None:
                data = base64.b64decode(binary_data)
        except binascii.Error as e:
            print(
                f'The Binary staging data for variable {variable} is not properly base64 encoded '
                f'due to {e}'
            )
            sys.exit()
        return data


class ThreatConnect:
    """Stages the ThreatConnect Data"""

    def __init__(self, provider):
        """Initialize Class properties."""
        self.provider = provider
        self.batch = None

    def dir(self, directory, owner, batch=False):
        """Stages the directory"""
        entities = []
        for stage_file in os.listdir(directory):
            if not (stage_file.endswith('.json') and stage_file.startswith('tc_stage_')):
                continue
            entities.append(self._convert_to_entities(f'{directory}/{stage_file}'))
        return self.entities(entities, owner, batch=batch)

    def file(self, file, owner, batch=False):
        """Stages the file"""
        entities = self._convert_to_entities(file)
        return self.entities(entities, owner, batch=batch)

    def entities(self, entities, owner=None, batch=False):
        """Stages all of the provided entities"""
        response = []
        if batch:
            self.batch = self.provider.tcex.batch(owner)
            for key, entity in entities.items():
                labels = entity.pop('securityLabels')
                attributes = entity.pop('attributes')
                tags = entity.pop('tags')
                associations = entity.pop('associations')
                entity['xid'] = entity.get(
                    'xid',
                    self.provider.batch.generate_xid(
                        [owner, entity.get('type'), entity.get('summary')]
                    ),
                )
                batch_entity = self._convert_to_batch_entity(entity)
                self.provider.batch.save(batch_entity)
                for tag in tags:
                    batch_entity.tag(tag)
                for label in labels:
                    batch_entity.label(label)
                for association in associations:
                    batch_entity.association(association)
                for attribute in attributes:
                    batch_entity.attribute(
                        attribute.get('type'),
                        attribute.get('value'),
                        attribute.get('displayed', True),
                    )

            response = self.batch.submit_all()
        else:
            for key, value in entities.items():
                response.append(self.entity(key, value, owner))
        return response if not batch else None

    def entity(self, key, value, owner=None):
        """Stage data in ThreatConnect"""
        owner = value.pop('owner', None) or owner
        created_entity = self.provider.tcex.cm.create_entity(value, owner)
        if created_entity is None:
            # This is here because there is a type `Task` in both TI and CM
            if value.get('type', '').lower().startswith('ti_'):
                value['type'] = value.get('type')[3:]
            created_entity = self.provider.tcex.ti.create_entity(value, owner)
        return {'key': key, 'data': created_entity}

    def delete_staged(self, staged_data):
        """Delete data in redis"""
        for data in staged_data:
            data = data.get('data', {})
            if data.get('status_code') != 201:
                continue

            entity_type = data.pop('main_type')
            ti = None
            if entity_type == 'Group':
                ti = self.provider.tcex.ti.group(data.get('sub_type'), owner=data.get('owner'))
                data = data.get(ti.api_entity) or data
                ti._set_unique_id(data)
            elif entity_type == 'Indicator':
                ti = self.provider.tcex.ti.indicator(data.get('sub_type'), owner=data.get('owner'))
                data = data.get(ti.api_entity) or data
                ti._set_unique_id(data)
            elif entity_type == 'Task':
                ti = self.provider.tcex.ti.task(owner=data.get('owner'))
                data = data.get(ti.api_entity) or data
                ti._set_unique_id(data)
            elif entity_type == 'Victim':
                ti = self.provider.tcex.ti.victim(owner=data.get('owner'))
                data = data.get(ti.api_entity) or data
                ti._set_unique_id(data)
            if ti:
                ti.delete()
            if entity_type == 'Case_Management':
                cm = getattr(self.provider.tcex.cm, data.get('sub_type'))()
                if data.get('sub_type').lower() in [
                    'workflow_event',
                    'workflowevent',
                    'workflow event',
                ]:
                    continue
                cm.id = data.get('id')
                cm.delete()

    def clear(self, owner):
        """Delete and recreate the owner"""

    @staticmethod
    def _convert_to_entities(file):
        """Convert A file to TC Entity's"""
        with open(file) as read_file:
            data = json.load(read_file)
        return data

    def _convert_to_batch_entity(self, entity):
        """Convert TC Entity to a Batch entity"""
        entity_type = entity.pop('type')
        ti = None
        if entity_type in self.provider.tcex.indicator_types:
            ti = self.provider.batch.indicator(entity_type, entity.pop('summary'), entity)
        elif entity_type in self.provider.tcex.group_types:
            ti = self.provider.batch.group(entity_type, entity.pop('summary'), entity)
        return ti



"""TcEx Framework Playbook module"""
# standard library
import re

from .playbooks_base import PlaybooksBase


class Playbooks(PlaybooksBase):
    """Playbook methods for accessing key value store.

    Args:
        tcex (object): Instance of TcEx.
        context (str): The KV Store context/session_id.
        output_variables (list): The requested output variables.
    """

    def __init__(self, tcex, context, output_variables):
        """Initialize the Class properties."""
        super().__init__(tcex, context, output_variables)

        # properties
        self.output_data = {}

    def add_output(self, key, value, variable_type, append_array=True):
        """Dynamically add output to output_data dictionary to be written to DB later.

        This method provides an alternative and more dynamic way to create output variables in an
        App. Instead of storing the output data manually and writing all at once the data can be
        stored inline, when it is generated and then written before the App completes.

        .. code-block:: python
            :linenos:
            :lineno-start: 1

            colors = ['blue', 'red', 'yellow']:
            tcex.playbook.add_output('app.colors', colors, 'StringArray')

            tcex.playbook.write_output()  #  writes the output stored in output_data

        .. code-block:: json
            :linenos:
            :lineno-start: 1

            {
                "my_color-String": {
                    "key": "my_color",
                    "type": "String",
                    "value": "blue"
                },
                "my_numbers-String": {
                    "key": "my_numbers",
                    "type": "String",
                    "value": "seven"
                },
                "my_numbers-StringArray": {
                    "key": "my_numbers",
                    "type": "StringArray",
                    "value": ["seven", "five"]
                }
            }

        Args:
            key (str): The variable name to write to storage.
            value (any): The value to write to storage.
            variable_type (str): The variable type being written.
            append_array (bool): If True arrays will be appended instead of being overwritten.

        """
        index = f'{key}-{variable_type}'
        self.output_data.setdefault(index, {})
        if value is None:
            return

        if variable_type in self._variable_array_types and append_array:
            self.output_data[index].setdefault('key', key)
            self.output_data[index].setdefault('type', variable_type)
            if isinstance(value, list):
                self.output_data[index].setdefault('value', []).extend(value)
            else:
                self.output_data[index].setdefault('value', []).append(value)
        else:
            self.output_data[index] = {'key': key, 'type': variable_type, 'value': value}

    def check_output_variable(self, variable):
        """Check to see if output variable was requested by downstream app.

        Using the auto generated dictionary of output variables check to see if provided
        variable was requested by downstream app.

        Args:
            variable (str): The variable name, not the full variable.

        Returns:
            (boolean): Boolean value indicator whether a match was found.
        """
        return variable in self.output_variables_by_name

    def create(self, key, value):
        """Create method of CRUD operation for working with KeyValue DB.

        This method will automatically determine the variable type and
        call the appropriate method to write the data.  If a non standard
        type is provided the data will be written as RAW data.

        Args:
            key (str): The variable to write to the DB.
            value (any): The data to write to the DB.

        Returns:
            (str): Result string of DB write.
        """
        data = None
        if key is not None:
            key = key.strip()
            parsed_key = self.parse_variable(key.strip())
            variable_type = parsed_key['type']

            # log/debug
            self.log.debug(f'create variable {key}')
            if variable_type not in ['Binary', 'BinaryArray']:
                self.log.trace(f'variable value: {value}')

            if variable_type in self._variable_single_types:
                data = self._create(key, value)
            elif variable_type in self._variable_array_types:
                data = self._create_array(key, value)
            else:
                data = self.create_raw(key, value)
        return data

    def create_binary(self, key, value):
        """Create method of CRUD operation for binary data.

        Args:
            key (str): The variable to write to the Key Value Store.
            value (any): The data to write to the Key Value Store.

        Returns:
            (str): Result of Key Value Store write.
        """
        supported_variable_type = 'Binary'
        if self.variable_type(key) != supported_variable_type:
            raise RuntimeError(f'The key provided ({key}) is not a {supported_variable_type} key.')
        return self._create(key, value)

    def create_binary_array(self, key, value):
        """Create method of CRUD operation for binary array data.

        Args:
            key (str): The variable to write to the Key Value Store.
            value (any): The data to write to the Key Value Store.

        Returns:
            (str): Result of Key Value Store write.
        """
        supported_variable_type = 'BinaryArray'
        if self.variable_type(key) != supported_variable_type:
            raise RuntimeError(f'The key provided ({key}) is not a {supported_variable_type} key.')
        return self._create_array(key, value)

    def create_key_value(self, key, value):
        """Create method of CRUD operation for key/value data.

        Args:
            key (str): The variable to write to the DB.
            value (any): The data to write to the DB.

        Returns:
            (str): Result of DB write
        """
        supported_variable_type = 'KeyValue'
        if self.variable_type(key) != supported_variable_type:
            raise RuntimeError(f'The key provided ({key}) is not a {supported_variable_type} key.')
        return self._create(key, value)

    def create_key_value_array(self, key, value):
        """Create method of CRUD operation for key/value array data.

        Args:
            key (str): The variable to write to the DB.
            value (any): The data to write to the DB.

        Returns:
            (str): Result of DB write.
        """
        supported_variable_type = 'KeyValueArray'
        if self.variable_type(key) != supported_variable_type:
            raise RuntimeError(f'The key provided ({key}) is not a {supported_variable_type} key.')
        return self._create_array(key, value)

    def create_string(self, key, value):
        """Create method of CRUD operation for string data.

        Args:
            key (str): The variable to write to the DB.
            value (any): The data to write to the DB.

        Returns:
            (str): Result of DB write.
        """
        supported_variable_type = 'String'
        if self.variable_type(key) != supported_variable_type:
            raise RuntimeError(f'The key provided ({key}) is not a {supported_variable_type} key.')
        return self._create(key, value)

    def create_string_array(self, key, value):
        """Create method of CRUD operation for string array data.

        Args:
            key (str): The variable to write to the DB.
            value (any): The data to write to the DB.

        Returns:
            (str): Result of DB write.
        """
        supported_variable_type = 'StringArray'
        if self.variable_type(key) != supported_variable_type:
            raise RuntimeError(f'The key provided ({key}) is not a {supported_variable_type} key.')
        return self._create_array(key, value)

    def create_tc_entity(self, key, value):
        """Create method of CRUD operation for TC entity data.

        Args:
            key (str): The variable to write to the DB.
            value (any): The data to write to the DB.

        Returns:
            (str): Result of DB write.
        """
        supported_variable_type = 'TCEntity'
        if self.variable_type(key) != supported_variable_type:
            raise RuntimeError(f'The key provided ({key}) is not a {supported_variable_type} key.')
        return self._create(key, value)

    def create_tc_entity_array(self, key, value):
        """Create method of CRUD operation for TC entity array data.

        Args:
            key (str): The variable to write to the DB.
            value (any): The data to write to the DB.

        Returns:
            (str): Result of DB write.
        """
        supported_variable_type = ['TCEntityArray', 'TCEnhancedEntity', 'TCEnhancedEntityArray']
        if self.variable_type(key) not in supported_variable_type:
            raise RuntimeError(f'The key provided ({key}) is not a {supported_variable_type} key.')
        return self._create_array(key, value)

    def create_output(self, key, value, variable_type=None):
        """Alias for Create method of CRUD operation for working with KeyValue DB.

        This method will automatically check to see if provided variable was requested by
        a downstream app and if so create the data in the KeyValue DB.

        Args:
            key (str): The variable to write to the DB.
            value (any): The data to write to the DB.
            variable_type (str): The variable type being written.

        Returns:
            (str): Result string of DB write.
        """
        #  This is if no downstream variables are requested then nothing should be returned.
        if not self.output_variables_by_type:  # pragma: no cover
            self.log.debug(f'Variable {key} was NOT requested by downstream app.')
            return None

        if key is None:
            self.log.info('Key has a none value and will not be written.')
            return None

        if value is None:
            self.log.info(f'Variable {key} has a none value and will not be written.')
            return None

        key = key.strip()
        key_type = f'{key}-{variable_type}'
        results = None
        if self.output_variables_by_type.get(key_type) is not None:
            # variable key-type has been requested
            v = self.output_variables_by_type.get(key_type)
            self.log.info(f"Variable {v.get('variable')} was requested by downstream App.")
            results = self.create(v.get('variable'), value)
        elif self.output_variables_by_name.get(key) is not None and variable_type is None:
            # variable key has been requested
            v = self.output_variables_by_name.get(key)
            self.log.info(f"Variable {v.get('variable')} was requested by downstream App.")
            results = self.create(v.get('variable'), value)
        else:
            self.log.trace(f'requested output variables: {self.output_variables_by_name}')
            self.log.debug(f'Variable {key} was NOT requested by downstream app.')

        return results

    def delete(self, key):
        """Delete method of CRUD operation for all data types.

        Args:
            key (str): The variable to write to the DB.

        Returns:
            (str): Result of DB write.
        """
        data = None
        if key is not None:
            data = self.key_value_store.delete(key.strip())
        else:  # pragma: no cover
            self.log.warning('The key field was None.')
        return data

    def exit(self, code=None, msg=None):
        """Exit the App

        Playbooks do not support partial failures so we change the exit method from 3 to 1 and call
        it a partial success instead.

        Args:
            code (int): The exit code value for the app.
            msg (str): A message to log and add to message tc output.
        """
        if code is None:
            code = self.tcex.exit_code
            if code == 3:
                self.log.info('Changing exit code from 3 to 0.')
                code = 0  # playbooks doesn't support partial failure
        elif code not in [0, 1]:
            code = 1

        # write outputs before exiting
        self.write_output()

        # required only for tcex testing framework
        if self.tcex.args.tcex_testing_context is not None:  # pragma: no cover
            self.tcex.redis_client.hset(self.tcex.args.tcex_testing_context, '_exit_message', msg)

        self.tcex.exit(code, msg)

    def is_variable(self, key):
        """Return True if provided key is a properly formatted variable."""
        if not isinstance(key, str):
            return False
        if re.match(self._variable_match, key):
            return True
        return False

    @property
    def output_variables_by_name(self):
        """Return output variables stored as name dict."""
        if self._output_variables_by_name is None:
            self._parse_output_variables()
        return self._output_variables_by_name

    @property
    def output_variables_by_type(self):
        """Return output variables stored as name-type dict."""
        if self._output_variables_by_type is None:
            self._parse_output_variables()
        return self._output_variables_by_type

    def parse_variable(self, variable):
        """Parse an input or output variable.

        **Example Variable**::

        #App:1234:output!String

        Args:
            variable (str): The variable name to parse.

        Returns:
            (dictionary): Result of parsed string.
        """
        data = None
        if variable is not None:
            variable = variable.strip()
            if re.match(self._variable_match, variable):
                var = re.search(self._variable_parse, variable)
                data = {
                    'root': var.group(0),
                    'job_id': var.group(2),
                    'name': var.group(3),
                    'type': var.group(4),
                }
        return data

    def read(self, key, array=False, embedded=True):
        """Read method of CRUD operation for working with KeyValue DB.

        This method will automatically check to see if a single variable is passed
        or if "mixed" data is passed and return the results from the DB. It will also
        automatically determine the variable type to read.

        Args:
            key (str): The variable to read from the DB.
            array (boolean): Convert string/dict to Array/List before returning.
            embedded (boolean): Resolve embedded variables.

        Returns:
            (any): Results retrieved from DB
        """
        # if a non-variable value is passed it should be the default
        value = key
        if isinstance(key, str):
            key = key.strip()
            variable_type = self.variable_type(key)
            if re.match(self._variable_match, key):
                # only log key if it's a variable
                self.log.debug(f'read variable {key}')
                if variable_type in self._variable_single_types:
                    value = self._read(key=key, embedded=embedded)
                elif variable_type in self._variable_array_types:
                    value = self._read_array(key=key, embedded=embedded)
                else:
                    value = self.read_raw(key)
            else:
                if variable_type == 'String':
                    # replace "\s" with a space only for user input.
                    # using '\\s' will prevent replacement.
                    value = re.sub(r'(?<!\\)\\s', ' ', value)
                    value = re.sub(r'\\\\s', r'\\s', value)

                if embedded:
                    # check for any embedded variables
                    value = self._read_embedded(value)

        # return data as a list
        if array and not isinstance(value, list):
            if value is not None:
                value = [value]
            else:
                # Adding none value to list breaks App logic. It's better to not request
                # Array and build array externally if None values are required.
                value = []

        return value

    def _entity_field(self, key, field, entity_type=None, default=None):
        """Read the value of the given key and return the data at the given field of the value.

        This method is used by functions designed to make it easier to get
        data from a particular field from TC(Enhanced)Entity(Array).

        Args:
            key (str): The variable to read from the DB.
            field (str): The field to find in the data.
            entity_type (Any): Default is None. The type of data being collected (valid values:
                ['groups', 'indicators']).
            default (Any): Default is None. The value to use for malformed TCEntities or
                TCEnhancedEntities.

        Returns:
            (list): A list of strings containing the desired values.
        """
        read_results = self.read(key, array=True)
        if not read_results:
            return []

        variable_type = self.variable_type(key).lower()
        enhanced_entity_field = None
        VALID_ENTITY_TYPES = ['group', 'indicator']

        # handle the odd format of tcEnhancedEntityArrays which follow the format detailed here:
        # https://docs.threatconnect.com/en/latest/rest_api/indicators/indicators.html#batch-indicator-input-file-format-v2
        if read_results and variable_type == 'tcenhancedentityarray':
            # type specific values
            if entity_type == 'indicator' and field == 'value':
                enhanced_entity_field = 'summary'
            elif entity_type == 'group' and field == 'value':
                enhanced_entity_field = 'name'
            elif entity_type not in VALID_ENTITY_TYPES:  # pragma: no cover
                message = (
                    f'Invalid entity_type ({entity_type}). Valid options: {VALID_ENTITY_TYPES}.'
                )
                raise RuntimeError(message)

            read_results = read_results[0].get(entity_type, [])

        is_tc_enhanced = enhanced_entity_field and variable_type in [
            'tcenhancedentity',
            'tcenhancedentityarray',
        ]
        is_tc_entity = variable_type in ['tcentity', 'tcentityarray']

        if is_tc_enhanced:
            values = [i.get(enhanced_entity_field, default) for i in read_results]
        elif is_tc_entity:
            values = [i.get(field, default) for i in read_results]
        else:
            values = read_results

        return values

    def read_indicator_values(self, key, default=None):
        """Read the value of the given key and return indicators from the value.

        This method will call the `read` method and then will process the data so as to return a
        list of strings where each string is an indicator (the summary of an indicator - e.g.
        ["foo.example.com", "bar.example.com"]).

        Args:
            key (str): The variable to read from the DB.
            default (Any): Default is None. The value to use for malformed TCEntities or
                TCEnhancedEntities.

        Returns:
            (list): A list of strings containing the indicators
        """
        return self._entity_field(key, 'value', entity_type='indicator', default=default)

    def read_group_values(self, key, default=None):
        """Read the value of the given key and return group names from the value.

        This method will call the `read` method and then will process the data
        so as to return a list of strings where each string is a group name.

        Args:
            key (str): The variable to read from the DB.
            default (Any): Default is None. The value to use for malformed TCEntities or
                TCEnhancedEntities.

        Returns:
            (list): A list of strings containing the group names
        """
        return self._entity_field(key, 'value', entity_type='group', default=default)

    def read_group_ids(self, key, default=None):
        """Read the value of the given key and return group ids from the value.

        This method will call the `read` method and then will process the data
        so as to return a list of strings where each string is a group id.

        Args:
            key (str): The variable to read from the DB.
            default (Any): Default is None. The value to use for malformed TCEntities or
                TCEnhancedEntities.

        Returns:
            (list): A list of strings containing the group ids
        """
        return self._entity_field(key, 'id', entity_type='group', default=default)

    def read_choice(self, key, alt_key):
        """Read method for choice inputs.

        Behavior:
        * If key is "-- Select --" return None.
        * If key is "-- Variable Input --" return resolved alt_key.
        * Else return resolved value for key.

        Args:
            key (str): The variable to read from KeyValue Store.
            alt_key (str): The alternate variable to read from the KeyValue Store.

        Returns:
            (None|str): Results retrieved from KeyValue Store
        """
        if key is None:
            return None

        if key.strip() == '-- Select --':
            return None

        if key.strip() == '-- Variable Input --':
            return self.read(alt_key)

        return self.read(key)

    def read_array(self, key, embedded=True):
        """Read playbook variable and return array for any variable type.

        Args:
            key (str): The variable to read from the DB.
            embedded (boolean): Resolve embedded variables.

        Returns:
            (any): Results retrieved from DB
        """
        return self.read(key, True, embedded)

    def read_binary(self, key, b64decode=True, decode=False):
        """Read method of CRUD operation for binary data.

        Args:
            key (str): The variable to read from the DB.
            b64decode (bool): If true the data will be base64 decoded.
            decode (bool): If true the data will be decoded to a String.

        Returns:
            (bytes|string): Results retrieved from DB.
        """
        return self._read(key, b64decode=b64decode, decode=decode)

    def read_binary_array(self, key, b64decode=True, decode=False):
        """Read method of CRUD operation for binary array data.

        Args:
            key (str): The variable to read from the DB.
            b64decode (bool): If true the data will be base64 decoded.
            decode (bool): If true the data will be decoded to a String.

        Returns:
            (list): Results retrieved from DB.
        """
        return self._read_array(key, b64decode=b64decode, decode=decode)

    def read_key_value(self, key, embedded=True):
        """Read method of CRUD operation for key/value data.

        Args:
            key (str): The variable to read from the DB.
            embedded (boolean): Resolve embedded variables.

        Returns:
            (dictionary): Results retrieved from DB.
        """
        return self._read(key, embedded=embedded)

    def read_key_value_array(self, key, embedded=True):
        """Read method of CRUD operation for key/value array data.

        Args:
            key (str): The variable to read from the DB.
            embedded (boolean): Resolve embedded variables.

        Returns:
            (list): Results retrieved from DB.
        """
        return self._read_array(key, embedded=embedded)

    def read_string(self, key, embedded=True):
        """Read method of CRUD operation for string data.

        Args:
            key (str): The variable to read from the DB.
            embedded (boolean): Resolve embedded variables.

        Returns:
            (str): Results retrieved from DB.
        """
        return self._read(key, embedded=embedded)

    def read_string_array(self, key, embedded=True):
        """Read method of CRUD operation for string array data.

        Args:
            key (str): The variable to read from the DB.
            embedded (boolean): Resolve embedded variables.

        Returns:
            (list): Results retrieved from DB.
        """
        return self._read_array(key, embedded=embedded)

    def read_tc_entity(self, key, embedded=True):
        """Read method of CRUD operation for TC entity data.

        Args:
            key (str): The variable to read from the DB.
            embedded (boolean): Resolve embedded variables.

        Returns:
            (dictionary): Results retrieved from DB.
        """
        return self._read(key, embedded=embedded)

    def read_tc_entity_array(self, key, embedded=True):
        """Read method of CRUD operation for TC entity array data.

        Args:
            key (str): The variable to read from the DB.
            embedded (boolean): Resolve embedded variables.

        Returns:
            (list): Results retrieved from DB.
        """
        return self._read_array(key, embedded=embedded)

    def variable_type(self, variable):
        """Get the Type from the variable string or default to String type.

        The default type is "String" for those cases when the input variable is
        contains not "DB variable" and is just a String.

        **Example Variable**::

            #App:1234:output!StringArray returns **StringArray**

        **Example String**::

            "My Data" returns **String**

        Args:
            variable (str): The variable to be parsed

        Returns:
            (str): The variable type.
        """
        var_type = 'String'
        if isinstance(variable, str):
            variable = variable.strip()
            if re.match(self._variable_match, variable):
                var_type = re.search(self._variable_parse, variable).group(4)
        return var_type

    def write_output(self):
        """Write all stored output data to storage."""
        for data in self.output_data.values():
            self.create_output(data.get('key'), data.get('value'), data.get('type'))



"""TcEx Service Common Module"""
# standard library
import http.server
import json
import os
import socketserver
import sys
import time
from threading import Event, Thread
from typing import Optional
from urllib.parse import parse_qs, urlparse
from uuid import uuid4

# first-party
from tcex.sessions import ExternalSession

from .test_case_service_common import TestCaseServiceCommon


class ApiServer(socketserver.ThreadingMixIn, http.server.HTTPServer):
    """HTTP Server for testing API Services"""

    allow_reuse_address = True

    def __init__(self, test_case: object, bind_addr: tuple):
        """Initialize Class properties

        Args:
            test_case: The test_case object.
            bind_addr: The binding address w/ port.
        """
        super().__init__(bind_addr, RequestHandler)
        self.test_case = test_case

        # properties
        self._host = bind_addr[0]
        self._port = bind_addr[1]
        self.active_requests = {}
        self.active_responses = {}
        self.args = test_case.default_args
        self.log = test_case.log
        self.message_broker = test_case.message_broker
        self.mqtt_client = test_case.message_broker.client

        # start server thread
        service = Thread(group=None, target=self.run, name='SimpleServerThread', daemon=True)
        service.start()

    def listen(self):
        """List for message coming from broker."""
        self.message_broker.add_on_connect_callback(self.on_connect)
        self.message_broker.add_on_message_callback(
            callback=self.on_message, topics=[self.test_case.client_topic]
        )

        t = Thread(name='broker-listener', target=self.message_broker.connect, args=(), daemon=True)
        t.start()

    def on_connect(self, client, userdata, flags, rc):  # pylint: disable=unused-argument
        """Handle message broker on_connect events."""
        # subscribe to client topic
        client.subscribe(self.test_case.client_topic)

    def on_message(self, client, userdata, message):  # pylint: disable=unused-argument
        """Handle message broker on_message events."""
        try:
            m = json.loads(message.payload)
        except ValueError:
            raise RuntimeError(f'Could not parse API service response JSON. ({message})')

        # only process RunService Acknowledged commands.
        if m.get('command').lower() == 'acknowledged' and m.get('type').lower() == 'runservice':
            self.active_responses[m.get('requestKey')] = m
            self.active_requests.pop(m.get('requestKey')).set()

    def run(self):
        """Run the server in threat."""
        print(f'\nRunning server: http://{self._host}:{self._port}')
        self.serve_forever()


class RequestHandler(http.server.BaseHTTPRequestHandler):
    """Request handler to forward request to API service."""

    def _build_request(self, method: str) -> dict:
        """Return request built from incoming HTTP request.

        {
            "apiToken": "SVC:5:RgIo6v:1596670377509:95:vWO1zu8W0a2NyXctWORKMe/kA616P6Vk8dsYvG ... ",
            "appId": 95,
            "bodyVariable": "request.body",
            "command": "RunService",
            "expireSeconds": 1596670377,
            "headers": [
                {
                    "name": "Accept",
                    "value": "*/*"
                },
                {
                    "name": "User-Agent",
                    "value": "PostmanRuntime/7.26.2"
                },
                {
                    "name": "Content-Type",
                    "value": "application/json"
                }
            ],
            "method": "GET",
            "path": "/data",
            "queryParams": [
                {
                    "name": "max",
                    "value": "1000"
                }
            ],
            "requestKey": "c29927c8-b94d-4116-a397-e6eb7002f41c"
        }

        Args:
            method: The HTTP method.

        Returns:
            dict: The response to send to API service over message broker topic.
        """
        url_parts = urlparse(self.path)

        # query params
        params = []
        for name, value in parse_qs(url_parts.query).items():
            if isinstance(value, list):
                for v in value:
                    params.append({'name': name, 'value': v})
            else:
                params.append({'name': name, 'value': value})

        # forward request to service
        request_key = str(uuid4())

        content_length = int(self.headers.get('content-length', 0))
        if content_length:
            body = self.rfile.read(content_length)
            self.server.test_case.redis_client.hset(request_key, 'request.body', body)

        request_url = self.headers.get('Host', 'http://localhost:8042')

        if request_url and not request_url.startswith(('http://', 'https://')):
            request_url = f'https://{request_url}'

        return {
            'apiToken': self.server.test_case.tc_token,
            'appId': 95,
            'bodyVariable': 'request.body',
            'command': 'RunService',
            'expireSeconds': int(time.time() + 600),
            'headers': [{'name': name, 'value': value} for name, value in self.headers.items()],
            'method': method,
            'path': url_parts.path,
            'queryParams': params,
            'requestKey': request_key,
            'requestUrl': request_url,
            'remoteAddress': '127.0.0.1',
        }

    def _build_response(self, response: Optional[dict] = None) -> None:
        """Build response data from API service response.

        {
            "bodyVariable": "response.body",
            "command": "Acknowledged",
            "headers": [
                {
                    "name": "x-cache",
                    "value": "MISS"
                },
                {
                    "name": "retry-after",
                    "value": "20"
                },
                {
                    "name": "content-type",
                    "value": "application/json"
                },
            ],
            "requestKey": "97190c5a-05e7-493d-8cb5-33844190eb72",
            "status": "Too Many Requests",
            "statusCode": "429",
            "type": "RunService"
        }

        Args:
            response: The response data from API service.
        """
        if response is None:
            self.send_error(500, message='No response sent on message broker client channel.')
            return

        # status code
        self.send_response(int(response.get('statusCode')))

        # headers
        for header in response.get('headers'):
            self.send_header(header.get('name'), str(header.get('value')))
        self.end_headers()

        # body
        body = self.server.test_case.redis_client.hget(response.get('requestKey'), 'response.body')
        if body is not None:
            self.wfile.write(body)

    def call_service(self, method: str):  # pylint: disable=useless-return
        """Call the API Service

        Args:
            method: The HTTP method.
        """
        request = self._build_request(method)
        request_key = request.get('requestKey')

        # create lock and sve request
        event = Event()
        self.server.active_requests[request_key] = event

        # publish run service
        self.server.test_case.publish(
            message=json.dumps(request), topic=self.server.test_case.server_topic
        )

        # block for x seconds
        event.wait(60)
        response: dict = self.server.active_responses.pop(request_key, None)

        self._build_response(response=response)

        return

    def do_DELETE(self):
        """Handle DELETE method."""
        return self.call_service('DELETE')

    def do_GET(self):
        """Handle GET method."""
        return self.call_service('GET')

    def do_PATCH(self):
        """Handle PATCH method."""
        return self.call_service('PATCH')

    def do_POST(self):
        """Handle POST method."""
        return self.call_service('POST')


class TestCaseApiService(TestCaseServiceCommon):
    """Service App TestCase Class"""

    api_service_host = os.getenv('API_SERVICE_HOST', 'localhost')
    api_service_port = int(os.getenv('API_SERVICE_PORT', '8042'))
    api_server = None
    stop_server = False

    def on_message(self, client, userdata, message):  # pylint: disable=unused-argument
        """Handle message broker on_message shutdown command events."""
        try:
            m = json.loads(message.payload)
        except ValueError:
            raise RuntimeError(f'Could not parse API service response JSON. ({message})')

        # only process RunService Acknowledged commands.
        if message.topic == self.server_topic and m.get('command').lower() == 'shutdown':
            self.stop_server = True

    def run(self):
        """Run the Playbook App.

        Returns:
            int: The App exit code
        """
        # first-party
        from run import run  # pylint: disable=no-name-in-module

        # backup sys.argv
        sys_argv_orig = sys.argv

        # clear sys.argv
        sys.argv = sys.argv[:1]

        # run the app
        exit_code = 0
        try:
            # provide callback to to run.py method on Trigger Service Apps
            run(set_app=self._app_callback)  # pylint: disable=unexpected-keyword-arg
        except SystemExit as e:
            exit_code = e.code

        # restore sys.argv
        sys.argv = sys_argv_orig

        self.log.data('run', 'Exit Code', exit_code)
        return exit_code

    def setup_method(self):
        """Run before each test method runs."""
        self.api_server = ApiServer(self, (self.api_service_host, self.api_service_port))
        self.api_server.listen()

        # subscribe to server topic
        self.message_broker.client.subscribe(self.server_topic)

        # register on_message shutdown monitor
        self.message_broker.add_on_message_callback(
            callback=self.on_message, index=0, topics=[self.server_topic]
        )

        super().setup_method()

    def teardown_method(self):
        """Run after each test method runs."""
        self.api_server.server_close()
        super().teardown_method()

    @property
    def test_client(self):
        """Return test client."""
        base_url = f'http://{self.api_service_host}:{self.api_service_port}'
        return ExternalSession(base_url)



"""TcEx Framework TcexJson Object."""
# standard library
import json
import logging
import os
import re
from collections import OrderedDict

# third-party
import colorama as c

from ..env_store import EnvStore
from .install_json import InstallJson


class TcexJson:
    """Object for tcex.json file.

    Args:
        filename (str, optional): The config filename. Defaults to tcex.json.
        path (str, optional): The path to the file. Defaults to os.getcwd().
        logger (logging.Logger, optional): A instance of Logger. Defaults to None.
    """

    def __init__(self, filename=None, path=None, logger=None):
        """Initialize class properties."""
        self._filename = filename or 'tcex.json'
        self._path = path or os.getcwd()
        self.log = logger or logging.getLogger('layout_json')

        # properties
        self._contents = None
        self.env_store = EnvStore()
        self.ij = InstallJson()

    @property
    def contents(self):
        """Return install.json contents."""
        if self._contents is None:
            try:
                with open(self.filename) as fh:
                    contents = json.load(fh, object_pairs_hook=OrderedDict)

                    # replace all variables
                    contents = self.replace_env_variables(contents)

                    # set updated contents
                    self._contents = contents
            except OSError:
                self._contents = {}

            # raise error if tcex.json is missing app_name field
            if self._contents and not self._contents.get('package', {}).get('app_name'):
                raise RuntimeError('The tcex.json file is missing the package.app_name field.')

            # log warning for old Apps
            if self._contents.get('package', {}).get('app_version'):
                print(
                    f'{c.Fore.YELLOW}'
                    'WARNING: The tcex.json file defines "app_version" which should only be '
                    'defined in legacy Apps. Removing the value can cause the App to be treated '
                    'as a new App by TcExchange. Please remove "app_version" when appropriate.'
                    f'{c.Fore.RESET}'
                )

        return self._contents

    @property
    def filename(self):
        """Return the fqpn for the layout.json file."""
        return os.path.join(self._path, self._filename)

    def replace_env_variables(self, json_data):
        """Update any env variable.

        Args:
            json_data (dict): The profile data dict.

        Returns:
            dict: The updated dict.
        """
        profile = json.dumps(json_data)

        for m in re.finditer(r'\${(env|envs|local|remote):(.*?)}', profile):
            try:
                full_match = m.group(0)
                env_type = m.group(1)
                env_key = m.group(2)

                env_value = self.env_store.getenv(env_key, env_type)
                if env_value is not None:
                    profile = profile.replace(full_match, env_value)
            except IndexError:
                print(f'{c.Fore.YELLOW}Invalid variable found {full_match}.')
        return json.loads(profile)

    def update(self, template=None):
        """Update the contents of the tcex.json file."""
        if not os.path.isfile(self.filename):
            return

        with open(self.filename, 'r+') as fh:
            json_data = json.load(fh)

            # update app_name
            self.update_package_app_name(json_data)

            # update deprecated fields
            self.update_deprecated_fields(json_data)

            # update package excludes
            self.update_package_excludes(json_data)

            # update package excludes
            self.update_lib_versions(json_data)

            # update variable pattern
            json_data = self.update_variable_pattern_env(json_data)

            # update template
            if template is not None:
                json_data['template'] = template

            # write updated profile
            fh.seek(0)
            fh.write(f'{json.dumps(json_data, indent=2, sort_keys=True)}\n')
            fh.truncate()

        # update contents
        self._contents = self.replace_env_variables(json_data)

    def update_package_app_name(self, json_data):
        """Update the package app_name in the tcex.json file."""
        if self.package_app_name in [None, '', 'TC_-_', 'TCPB_-_', 'TCVC_-_', 'TCVW_-_']:
            # lower case name and replace prefix if already exists
            app_name = os.path.basename(os.getcwd()).lower().replace(self.ij.app_prefix.lower(), '')

            # replace spaces and dashes with underscores
            app_name = app_name.replace(' ', '_').replace('-', '_').lower()

            # title case app name
            app_name = '_'.join([a.title() for a in app_name.split('_')])

            # prepend appropriate App prefix (e.g., TCPB_-_)
            app_name = f'{self.ij.app_prefix}{app_name}'

            # update App name
            json_data['package']['app_name'] = app_name

    @staticmethod
    def update_deprecated_fields(json_data):
        """Update deprecated fields in the tcex.json file."""
        deprecated_fields = ['profile_include_dirs']
        for d in deprecated_fields:
            try:
                del json_data[d]
            except KeyError:
                pass

    def update_package_excludes(self, json_data):
        """Update the excludes values in the tcex.json file."""
        excludes = self.package_excludes
        excludes.extend(
            [
                '.gitignore',
                '.pre-commit-config.yaml',
                'local-*',
                'pyproject.toml',
                'setup.cfg',
                'tcex.json',
                '*.install.json',
                'tcex.d',
            ]
        )
        excludes = sorted(list(set(excludes)))
        # the requirements.txt file is required for App Builder
        try:
            excludes.remove('requirements.txt')
        except ValueError:
            pass
        json_data['package']['excludes'] = excludes

    def update_lib_versions(self, json_data):
        """Update the lib_versions array in the tcex.json file."""
        if os.getenv('TCEX_LIB_VERSIONS') and not self.lib_versions:
            lib_versions = []
            for version in os.getenv('TCEX_LIB_VERSIONS').split(','):
                lib_versions.append(
                    {
                        'lib_dir': f'lib_${{env:{version}}}',
                        'python_executable': f'~/.pyenv/versions/${{env:{version}}}/bin/python',
                    }
                )
            json_data['lib_versions'] = lib_versions

    @staticmethod
    def update_variable_pattern_env(json_data):
        """Update the profile variable to latest pattern

        Args:
            json_data (dict): The profile data dict.

        Returns:
            dict: The updated dict.
        """
        data = json.dumps(json_data)

        for m in re.finditer(r'\$(env|envs|local|remote)\.([^(\/|\")]*)', data):
            try:
                full_match = m.group(0)
                env_type = m.group(1)
                env_key = m.group(2)

                new_variable = f'${{{env_type}:{env_key}}}'
                data = data.replace(full_match, new_variable)
            except IndexError:
                print(f'{c.Fore.YELLOW}Invalid variable found {full_match}.')
        return json.loads(data)

    #
    # properties
    #

    @property
    def lib_versions(self):
        """Return property."""
        return self.contents.get('lib_versions', [])

    @property
    def package(self):
        """Return property."""
        return self.contents.get('package', {})

    @property
    def package_app_name(self):
        """Return property."""
        return self.package.get('app_name')

    @property
    def package_app_version(self):
        """Return property."""
        return self.package.get('app_version')

    @property
    def package_bundle(self):
        """Return property."""
        return self.package.get('bundle', False)

    @property
    def package_bundle_name(self):
        """Return property."""
        return self.package.get('bundle_name')

    @property
    def package_bundle_packages(self):
        """Return property."""
        return self.package.get('bundle_packages', [])

    @property
    def package_excludes(self):
        """Return property."""
        return self.package.get('excludes', [])

    @property
    def template(self):
        """Return property."""
        return self.contents.get('template')



"""TcEx Framework InstallJson Object."""
# standard library
import json
import logging
import os
import uuid
from collections import OrderedDict


class InstallJson:
    """Object for install.json file.

    Args:
        filename (str, optional): The config filename. Defaults to install.json.
        path (str, optional): The path to the file. Defaults to os.getcwd().
        logger (logging.Logger, optional): A instance of Logger. Defaults to None.
    """

    def __init__(self, filename=None, path=None, logger=None):
        """Initialize class properties."""
        self._filename = filename or 'install.json'
        self._path = path or os.getcwd()
        self.log = logger or logging.getLogger('install_json')

        # properties
        self._contents = None

    @property
    def _commit_hash(self):
        """Return the current commit hash if available.

        This is not a required task so best effort is fine. In other words this is not guaranteed
        to work 100% of the time.
        """
        commit_hash = None
        branch = None
        branch_file = '.git/HEAD'  # ref: refs/heads/develop

        # get current branch
        if os.path.isfile(branch_file):
            with open(branch_file) as f:
                try:
                    branch = f.read().strip().split('/')[2]
                except IndexError:
                    pass

            # get commit hash
            if branch:
                hash_file = f'.git/refs/heads/{branch}'
                if os.path.isfile(hash_file):
                    with open(hash_file) as f:
                        commit_hash = f.read().strip()
        return commit_hash

    @staticmethod
    def _to_bool(value):
        """Convert string value to bool."""
        bool_value = False
        if str(value).lower() in ['1', 'true']:
            bool_value = True
        return bool_value

    @property
    def app_output_var_type(self):
        """Return the appropriate output var type for the current App."""
        if self.runtime_level.lower() in ['triggerservice', 'webhooktriggerservice']:
            return 'Trigger'
        return 'App'

    @property
    def app_prefix(self):
        """Return the appropriate output var type for the current App."""
        if self.runtime_level.lower() == 'organization':
            return 'TC_-_'
        if self.runtime_level.lower() == 'playbook':
            return 'TCPB_-_'
        if self.runtime_level.lower() == 'triggerservice':
            return 'TCVC_-_'
        if self.runtime_level.lower() == 'webhooktriggerservice':
            return 'TCVW-_'
        return ''

    @property
    def contents(self):
        """Return install.json contents."""
        if self._contents is None:
            try:
                with open(self.filename) as fh:
                    self._contents = json.load(fh, object_pairs_hook=OrderedDict)
            except OSError:
                self._contents = {'runtimeLevel': 'external'}
        return self._contents

    def create_output_variables(self, output_variables, job_id=9876):
        """Create output variables.

        # "#App:9876:app.data.count!String"
        # "#Trigger:9876:app.data.count!String"

        Args:
            output_variables (dict): A dict of the output variables
            job_id (int): A job id to use in output variable string.
        """
        variables = []
        for p in output_variables:
            variables.append(self.create_variable(p.get('name'), p.get('type'), job_id))
        return variables

    def create_variable(self, var_name, var_type, job_id=1234):
        """Create output variables.

        # "#App:9876:app.data.count!String"
        # "#Trigger:9876:app.data.count!String"

        Args:
            var_name (str): The variable name.
            var_type (str): The variable type.
            job_id (int): A job id to use in output variable string.
        """
        return f'#{self.app_output_var_type}:{job_id}:{var_name}!{var_type}'

    @staticmethod
    def expand_valid_values(valid_values):
        """Expand supported playbook variables to their full list.

        Args:
            valid_values (list): The list of valid values for Choice or MultiChoice inputs.

        Returns:
            list: An expanded list of valid values for Choice or MultiChoice inputs.
        """
        valid_values = list(valid_values)
        if '${GROUP_TYPES}' in valid_values:
            valid_values.remove('${GROUP_TYPES}')
            valid_values.extend(
                [
                    'Adversary',
                    'Campaign',
                    'Document',
                    'Email',
                    'Event',
                    'Incident',
                    'Intrusion Set',
                    'Signature',
                    'Task',
                    'Threat',
                ]
            )
        elif '${OWNERS}' in valid_values:
            valid_values.remove('${OWNERS}')
        elif '${USERS}' in valid_values:
            valid_values.remove('${USERS}')
        return valid_values

    @property
    def filename(self):
        """Return the fqpn for the layout.json file."""
        return os.path.join(self._path, self._filename)

    def filter_params_dict(
        self,
        name=None,
        hidden=None,
        required=None,
        service_config=None,
        _type=None,
        input_permutations=None,
    ):
        """Return params as name/data dict.

        Args:
            name (str, optional): The name of the input to return. Defaults to None.
            hidden (bool, optional): If set the inputs will be filtered based on hidden field.
            required (bool, optional): If set the inputs will be filtered based on required field.
            service_config (bool, optional): If set the inputs will be filtered based on
                serviceConfig field.
            _type (str, optional): The type of input to return. Defaults to None.
            input_permutations (list, optional): A list of valid input names for provided
                permutation.

        Returns:
            dict: All valid inputs for current filter.
        """
        params = {}
        for p in self.params:

            if name is not None:
                if p.get('name') != name:
                    continue

            if hidden is not None:
                if p.get('hidden', False) is not hidden:
                    continue

            if required is not None:
                if p.get('required', False) is not required:
                    continue

            if service_config is not None:
                if p.get('serviceConfig', False) is not service_config:
                    continue

            if _type is not None:
                if p.get('type') != _type:
                    continue

            if input_permutations is not None:
                if p.get('name') not in input_permutations:
                    continue

            params.setdefault(p.get('name'), p)
        return params

    def has_feature(self, feature):
        """Return True if App has the provided feature."""
        return feature.lower() in [f.lower() for f in self.features]

    @property
    def optional_params_dict(self):
        """Return params as name/data dict."""
        params = {}
        for p in self.params:
            if p.get('required', False) is False:
                params.setdefault(p.get('name'), p)
        return params

    @property
    def output_dict(self):
        """Return outputs as name/data dict."""
        outputs = {}
        for o in self.output_variables:
            outputs.setdefault(o.get('name'), o)
        return outputs

    @property
    def params_dict(self):
        """Return params as name/data dict."""
        params = {}
        for p in self.params:
            params.setdefault(p.get('name'), p)
        return params

    def params_to_args(
        self,
        name=None,
        hidden=None,
        required=None,
        service_config=None,
        _type=None,
        input_permutations=None,
    ):
        """Return params as cli args.

        Args:
            name (str, optional): The name of the input to return. Defaults to None.
            required (bool, optional): If set the inputs will be filtered based on required field.
            service_config (bool, optional): If set the inputs will be filtered based on
                serviceConfig field.
            _type (str, optional): The type of input to return. Defaults to None.
            input_permutations (list, optional): A list of valid input names for provided
                permutation.

        Returns:
            dict: All valid inputs for current filter.
        """
        args = {}
        for n, p in self.filter_params_dict(
            name, hidden, required, service_config, _type, input_permutations
        ).items():
            if p.get('type').lower() == 'boolean':
                args[n] = self._to_bool(p.get('default', False))
            elif p.get('type').lower() == 'choice':
                # use the value from input_permutations if available or provide valid values
                valid_values = f"[{'|'.join(self.expand_valid_values(p.get('validValues', [])))}]"
                if input_permutations is not None:
                    valid_values = input_permutations.get(n, valid_values)
                args[n] = valid_values
            elif p.get('type').lower() == 'multichoice':
                args[n] = p.get('validValues', [])
            elif p.get('type').lower() == 'keyvaluelist':
                args[n] = '<KeyValueArray>'
            elif n in ['api_access_id', 'api_secret_key']:
                # leave these parameters set to the value defined in defaults
                pass
            else:
                types = '|'.join(p.get('playbookDataType', []))
                if types:
                    args[n] = p.get('default', f'<{types}>')
                else:
                    args[n] = p.get('default', '')
        return args

    @property
    def require_params_dict(self):
        """Return params as name/data dict."""
        params = {}
        for p in self.params:
            if p.get('required') is True:
                params.setdefault(p.get('name'), p)
        return params

    @property
    def service_config_params_dict(self):
        """Return params as name/data dict."""
        params = {}
        for p in self.params:
            if p.get('serviceConfig') is True:
                params.setdefault(p.get('name'), p)
        return params

    @property
    def tc_playbook_out_variables(self):
        """Return playbook output variable name array"""
        return self.create_output_variables(self.output_variables)

    @property
    def tc_playbook_out_variables_csv(self):
        """Return playbook output variables as CSV string"""
        return ','.join(self.tc_playbook_out_variables)

    @property
    def trigger_config_params_dict(self):
        """Return params as name/data dict."""
        params = {}
        for p in self.params:
            if p.get('serviceConfig', False) is False:
                params.setdefault(p.get('name'), p)
        return params

    def update(
        self,
        commit_hash=False,
        features=True,
        migrate=False,
        sequence=True,
        valid_values=True,
        playbook_data_types=True,
    ):
        """Update the profile with all required changes.

        Args:
            commit_hash (bool, optional): If True the commitHash value will be added.
            features (bool, optional): If True the features array will be updated.
            migrate (bool, optional): If True the programMain will be set to "run".
            sequence (bool, optional): If True the sequence numbers will be updated.
            valid_values (bool, optional): If True the valid values will be updated.
            playbook_data_types (bool, optional):  If True the pbDataTypes will be updated.
        """
        with open(self.filename, 'r+') as fh:
            json_data = json.load(fh)

            # update appId field
            self.update_app_id(json_data)

            # update commitHash field
            if commit_hash is True:
                self.update_commit_hash(json_data)

            # update displayName field
            self.update_display_name(json_data)

            # update features array
            if features is True:
                self.update_features(json_data)

            if migrate is True:
                # update programMain to run
                self.update_program_main(json_data)

            # update sequence numbers
            if sequence is True:
                self.update_sequence_numbers(json_data)

            # update valid values
            if valid_values is True:
                self.update_valid_values(json_data)

            # update playbook data types
            if playbook_data_types is True:
                self.update_playbook_data_types(json_data)

            # write updated profile
            fh.seek(0)
            fh.write(f'{json.dumps(json_data, indent=2, sort_keys=True)}\n')
            fh.truncate()

        self._contents = json_data

    @staticmethod
    def update_app_id(json_data: dict) -> None:
        """Update to ensure an appId field exists.

        All App should have an appId to uniquely identify the App. this is not intended to be
        used by core to identify an App.  using appId + major Version could be used for unique
        App identification in core in a future release.
        """
        if json_data.get('appId') is None:
            json_data['appId'] = str(
                uuid.uuid5(uuid.NAMESPACE_X500, os.path.basename(os.getcwd()).lower())
            )

    def update_commit_hash(self, json_data: dict) -> None:
        """Update to ensure an appId field exists.

        Add/Update the commit hash to the install.json file if possible.
        """
        if self._commit_hash:
            json_data['commitHash'] = self._commit_hash

    def update_display_name(self, json_data: dict) -> None:
        """Update the displayName parameter."""
        if not json_data.get('displayName'):
            display_name = os.path.basename(os.getcwd()).replace(self.app_prefix, '')
            display_name = display_name.replace('_', ' ').replace('-', ' ')
            display_name = ' '.join([a.title() for a in display_name.split(' ')])
            json_data['displayName'] = display_name

    def update_features(self, json_data: dict) -> None:
        """Update feature set based on App type."""
        features = self.features
        if self.runtime_level.lower() in ['organization']:
            features = ['fileParams', 'secureParams']
        elif self.runtime_level.lower() in ['playbook']:
            features = [
                'aotExecutionEnabled',
                'appBuilderCompliant',
                'fileParams',
                'secureParams',
            ]
        elif self.runtime_level.lower() in [
            'apiservice',
            'triggerservice',
            'webhooktriggerservice',
        ]:
            features = ['appBuilderCompliant', 'fileParams']

        # add layoutEnabledApp if layout.json file exists in project
        if os.path.isfile(os.path.join(self._path, 'layout.json')):
            features.append('layoutEnabledApp')

        # re-add supported optional features
        for feature in self.features:
            if feature in [
                'advancedRequest',
                'CALSettings',
                'webhookResponseMarshall',
                'webhookServiceEndpoint',
            ]:
                features.append(feature)

        json_data['features'] = sorted(features)

    def update_program_main(self, json_data: dict) -> None:
        """Update program main on App type."""
        if self.program_main:
            if self.runtime_level.lower() in ['playbook']:
                json_data['programMain'] = 'run'
            elif self.runtime_level.lower() in ['triggerservice', 'webhooktriggerservice']:
                json_data['programMain'] = 'run'

    @staticmethod
    def update_sequence_numbers(json_data: dict) -> None:
        """Update program main on App type."""
        sequence_number = 1
        for param in json_data.get('params', []):
            param['sequence'] = sequence_number
            sequence_number += 1

    @staticmethod
    def update_valid_values(json_data: dict) -> None:
        """Update program main on App type."""
        for param in json_data.get('params', []):
            if param.get('type', None) not in ['String', 'KeyValueList']:
                continue
            if param.get('encrypt', False):
                if '${KEYCHAIN}' not in param.get('validValues', []):
                    param['validValues'] = param.get('validValues') or []
                    param['validValues'].append('${KEYCHAIN}')
            else:
                if '${TEXT}' not in (param.get('validValues') or []):
                    param['validValues'] = param.get('validValues') or []
                    param['validValues'].append('${TEXT}')

    @staticmethod
    def update_playbook_data_types(json_data: dict) -> None:
        """Update program main on App type."""
        if json_data.get('runtimeLevel', None) != 'Playbook':
            return

        for param in json_data.get('params', []):
            if param.get('type') != 'String':
                continue
            if param.get('playbookDataType') in [None, []]:
                param.setdefault('playbookDataType', []).append('String')

    def validate(self):
        """Validate install.json."""

        # check for duplicate input names
        self.validate_duplicate_input_names()

        # check for duplicate input names
        self.validate_duplicate_sequences()

    def validate_duplicate_input_names(self):
        """Check for duplicate input names."""
        duplicates = []
        name_tracker = []
        for p in self.params:
            name = p.get('name')
            if name in name_tracker:
                duplicates.append(name)
            name_tracker.append(name)
        return duplicates

    def validate_duplicate_sequences(self):
        """Check for duplicate sequence numbers."""
        duplicates = []
        sequence_tracker = []
        for p in self.params:
            sequence = p.get('sequence')
            if sequence in sequence_tracker:
                duplicates.append(sequence)
            sequence_tracker.append(sequence)
        return duplicates

    def validate_duplicate_outputs(self):
        """Check for duplicate input names."""
        duplicates = []
        name_type_tracker = []
        for o in self.output_variables:
            name = o.get('name')
            type_ = o.get('type')
            name_type = f'{name}-{type_}'
            if name_type in name_type_tracker:
                duplicates.append(name)
            name_type_tracker.append(name_type)
        return duplicates

    #
    # properties
    #

    @property
    def allow_on_demand(self):
        """Return property."""
        return self.contents.get('allowOnDemand', False)

    @property
    def commit_hash(self):
        """Return property."""
        return self.contents.get('commitHash')

    @property
    def display_name(self):
        """Return property."""
        return self.contents.get('displayName')

    @property
    def features(self):
        """Return property."""
        return self.contents.get('features', [])

    @property
    def feeds(self):
        """Return property."""
        return self.contents.get('feeds', {})

    @property
    def labels(self):
        """Return property."""
        return self.contents.get('labels', [])

    @property
    def language_version(self):
        """Return property."""
        return self.contents.get('languageVersion')

    @property
    def list_delimiter(self):
        """Return property."""
        return self.contents.get('listDelimiter')

    @property
    def min_server_version(self):
        """Return property."""
        return self.contents.get('minServerVersion')

    @property
    def note(self):
        """Return property."""
        return self.contents.get('note')

    @property
    def output_prefix(self):
        """Return output prefix."""
        return self.playbook.get('outputPrefix')

    @property
    def output_variables(self):
        """Return output variables dict."""
        return self.playbook.get('outputVariables', [])

    @property
    def params(self):
        """Return property."""
        return self.contents.get('params', [])

    @property
    def playbook(self):
        """Return property."""
        return self.contents.get('playbook', {})

    @property
    def playbook_type(self):
        """Return property."""
        return self.playbook.get('type')

    @property
    def program_icon(self):
        """Return property."""
        return self.contents.get('programIcon')

    @property
    def program_language(self):
        """Return property."""
        return self.contents.get('programLanguage')

    @property
    def program_main(self):
        """Return property."""
        return self.contents.get('programMain')

    @property
    def program_name(self):
        """Return property."""
        return self.contents.get('programName')

    @property
    def program_version(self):
        """Return property."""
        return self.contents.get('programVersion', '1.0.0')

    @property
    def publish_out_files(self):
        """Return property."""
        return self.contents.get('publishOutFiles')

    @property
    def repeating_minutes(self):
        """Return property."""
        return self.contents.get('repeatingMinutes')

    @property
    def runtime_context(self):
        """Return property."""
        return self.contents.get('runtimeContext')

    @property
    def runtime_level(self):
        """Return property."""
        return self.contents.get('runtimeLevel')

    @property
    def service(self):
        """Return service."""
        return self.contents.get('service', {})

    @property
    def service_discovery_types(self):
        """Return service."""
        return self.service.get('discoveryTypes', [])



"""TcEx Notification Module"""
# standard library
import json


class Notifications:
    """TcEx Notification Class"""

    def __init__(self, tcex):
        """Initialize the Class properties.

        Args:
            tcex (obj): An instance of TcEx object.
        """
        self.tcex = tcex

        # properties
        self._is_organization = False
        self._notification_type = None
        self._recipients = None
        self._priority = 'Low'

    def recipients(self, notification_type, recipients, priority='Low'):
        """Set vars for the passed in data. Used for one or more recipient notification.

        .. code-block:: javascript

            {
                "notificationType": notification_type,
                "priority": priority
                "isOrganization": false,
                "recipients": recipients
            }

        Args:
            notification_type (str): The type of notification being sent.
            recipients (str): A comma delimited string of recipients.
            priority (str): The priority: Low, Medium, High.
        """
        self._notification_type = notification_type
        self._recipients = recipients
        self._priority = priority
        self._is_organization = False

    def org(self, notification_type, priority='Low'):
        """Set vars for the passed in data. Used for org notification.

        .. code-block:: javascript

            {
                "notificationType": notification_type,
                "priority": priority
                "isOrganization": true
            }

        Args:
            notification_type (str): The notification type.
            priority (str): The priority: Low, Medium, High.
        """
        self._notification_type = notification_type
        self._recipients = None
        self._priority = priority
        self._is_organization = True

    def send(self, message):
        """Send our message

        Args:
            message (str): The message to be sent.

        Returns:
            requests.models.Response: The response from the request.

        """
        body = {
            'notificationType': self._notification_type,
            'priority': self._priority,
            'isOrganization': self._is_organization,
            'message': message,
        }

        if self._recipients:
            body['recipients'] = self._recipients

        self.tcex.log.debug(f'notification body: {json.dumps(body)}')

        # create our tcex resource
        r = self.tcex.session.post('/v2/notifications', json=body)
        if r.status_code == 400:
            # specifically handle unknown users
            self.tcex.log.error(f'Failed to send notification ({r.text})')
        elif not r.ok:  # pragma: no cover
            self.tcex.handle_error(750, [r.status_code, r.text])

        # return response body
        return r.json()
